import re

from pydantic import Field, SerializeAsAny, model_validator

from metagpt.actions.action import Action
from metagpt.environment.werewolf.const import RoleState, RoleType
from metagpt.ext.werewolf.actions import (
    ACTIONS,
    AddNewExperiences,
    InstructSpeak,
    NighttimeWhispers,
    Reflect,
    RetrieveExperiences,
    Speak,
)
from metagpt.ext.werewolf.schema import RoleExperience, WwMessage
from metagpt.logs import logger
from metagpt.roles import Role
from metagpt.utils.common import any_to_str


class BasePlayer(Role):
    name: str = "PlayerXYZ"
    profile: str = "BasePlayer"
    special_action_names: list[str] = []
    use_reflection: bool = True
    use_experience: bool = False
    use_memory_selection: bool = False
    new_experience_version: str = ""
    status: RoleState = RoleState.ALIVE

    special_actions: list[SerializeAsAny[Action]] = Field(default=[], validate_default=True)
    experiences: list[RoleExperience] = []

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # 技能和监听配置
        self._watch([InstructSpeak])  # 监听Moderator的指令以做行动
        special_actions = [ACTIONS[action_name] for action_name in self.special_action_names]
        capable_actions = [Speak] + special_actions
        self.set_actions(capable_actions)  # 给角色赋予行动技能
        self.special_actions = special_actions

        if not self.use_reflection and self.use_experience:
            logger.warning("You must enable use_reflection before using experience")
            self.use_experience = False

    @model_validator(mode="after")
    def check_addresses(self):
        if not self.addresses:
            self.addresses = {any_to_str(self), self.name, self.profile} if self.name else {any_to_str(self)}
        return self

    async def _observe(self, ignore_memory=False) -> int:
        if self.status != RoleState.ALIVE:
            # 死者不再参与游戏
            return 0
        news = []
        if not news:
            news = self.rc.msg_buffer.pop_all()
        old_messages = [] if ignore_memory else self.rc.memory.get()
        for m in news:
            if len(m.restricted_to) and self.profile not in m.restricted_to and self.name not in m.restricted_to:
                # if the msg is not send to the whole audience ("") nor this role (self.profile or self.name),
                # then this role should not be able to receive it and record it into its memory
                continue
            self.rc.memory.add(m)
        self.rc.news = [
            n for n in news if (n.cause_by in self.rc.watch or self.profile in n.send_to) and n not in old_messages
        ]

        # TODO to delete
        # await super()._observe()
        # # 只有发给全体的（""）或发给自己的（self.profile）消息需要走下面的_react流程，
        # # 其他的收听到即可，不用做动作
        # self.rc.news = [msg for msg in self.rc.news if msg.send_to in ["", self.profile]]
        return len(self.rc.news)

    async def _think(self):
        news = self.rc.news[0]
        assert news.cause_by == any_to_str(InstructSpeak)  # 消息为来自Moderator的指令时，才去做动作
        if not news.restricted_to:
            # 消息接收范围为全体角色的，做公开发言（发表投票观点也算发言）
            self.rc.todo = Speak()
        elif self.profile in news.restricted_to:
            # FIXME: hard code to split, restricted为"Moderator"或"Moderator, 角色profile"
            # Moderator加密发给自己的，意味着要执行角色的特殊动作
            self.rc.todo = self.special_actions[0]()
        return True

    async def _act(self):
        # todo为_think时确定的，有两种情况，Speak或Protect
        todo = self.rc.todo
        logger.info(f"{self._setting}: ready to {str(todo)}")

        # 可以用这个函数获取该角色的全部记忆和最新的instruction
        memories = self.get_all_memories()
        latest_instruction = self.get_latest_instruction()

        reflection = (
            await Reflect().run(
                profile=self.profile, name=self.name, context=memories, latest_instruction=latest_instruction
            )
            if self.use_reflection
            else ""
        )

        experiences = (
            RetrieveExperiences().run(
                query=reflection, profile=self.profile, excluded_version=self.new_experience_version
            )
            if self.use_experience
            else ""
        )

        # 根据自己定义的角色Action，对应地去run，run的入参可能不同
        if isinstance(todo, Speak):
            rsp = await todo.run(
                profile=self.profile,
                name=self.name,
                context=memories,
                latest_instruction=latest_instruction,
                reflection=reflection,
                experiences=experiences,
            )
            restricted_to = set()

        elif isinstance(todo, NighttimeWhispers):
            rsp = await todo.run(
                profile=self.profile, name=self.name, context=memories, reflection=reflection, experiences=experiences
            )
            restricted_to = {RoleType.MODERATOR.value, self.profile}  # 给Moderator发送使用特殊技能的加密消息
        msg = WwMessage(
            content=rsp,
            role=self.profile,
            sent_from=self.name,
            cause_by=type(todo),
            send_to={},
            restricted_to=restricted_to,
        )

        self.experiences.append(
            RoleExperience(
                name=self.name,
                profile=self.profile,
                reflection=reflection,
                instruction=latest_instruction,
                response=rsp,
                version=self.new_experience_version,
            )
        )

        logger.info(f"{self._setting}: {rsp}")

        return msg

    def get_all_memories(self) -> str:
        memories = self.rc.memory.get()
        time_stamp_pattern = r"[0-9]+ \| "
        # NOTE: 除Moderator外，其他角色使用memory，只能用m.sent_from（玩家名）不能用m.role（玩家角色），因为他们不知道说话者的身份
        memories = [f"{m.sent_from}: {re.sub(time_stamp_pattern, '', m.content)}" for m in memories]  # regex去掉时间戳
        memories = "\n".join(memories)
        return memories

    def get_latest_instruction(self) -> str:
        return self.rc.important_memory[-1].content  # 角色监听着Moderator的InstructSpeak，是其重要记忆，直接获取即可

    def set_status(self, new_status: RoleState):
        self.status = new_status

    def record_experiences(self, round_id: str, outcome: str, game_setup: str):
        experiences = [exp for exp in self.experiences if len(exp.reflection) > 2]  # not "" or not '""'
        for exp in experiences:
            exp.round_id = round_id
            exp.outcome = outcome
            exp.game_setup = game_setup
        AddNewExperiences().run(experiences)



from metagpt.environment.werewolf.const import RoleType
from metagpt.ext.werewolf.actions import InstructSpeak, Poison, Save, Speak
from metagpt.ext.werewolf.roles.base_player import BasePlayer
from metagpt.utils.common import any_to_str


class Witch(BasePlayer):
    name: str = RoleType.WITCH.value
    profile: str = RoleType.WITCH.value
    special_action_names: list[str] = ["Save", "Poison"]

    async def _think(self):
        """女巫涉及两个特殊技能，因此在此需要改写_think进行路由"""
        news = self.rc.news[0]
        assert news.cause_by == any_to_str(InstructSpeak)  # 消息为来自Moderator的指令时，才去做动作
        if not news.restricted_to:
            # 消息接收范围为全体角色的，做公开发言（发表投票观点也算发言）
            self.rc.todo = Speak()
        elif self.profile in news.restricted_to:
            # FIXME: hard code to split, restricted为"Moderator"或"Moderator,角色profile"
            # Moderator加密发给自己的，意味着要执行角色的特殊动作
            # 这里用关键词进行动作的选择，需要Moderator侧的指令进行配合
            if "save" in news.content.lower():
                self.rc.todo = Save()
            elif "poison" in news.content.lower():
                self.rc.todo = Poison()
            else:
                raise ValueError("Moderator's instructions must include save or poison keyword")
        return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Desc   : StanfordTown Action
import json
import time
from abc import abstractmethod
from pathlib import Path
from typing import Any, Optional, Union

from metagpt.actions.action import Action
from metagpt.config2 import config
from metagpt.ext.stanford_town.utils.const import PROMPTS_DIR
from metagpt.logs import logger


class STAction(Action):
    name: str = "STAction"
    prompt_dir: Path = PROMPTS_DIR
    fail_default_resp: Optional[str] = None

    @property
    def cls_name(self):
        return self.__class__.__name__

    @abstractmethod
    def _func_validate(self, llm_resp: str, prompt: str):
        raise NotImplementedError

    @abstractmethod
    def _func_cleanup(self, llm_resp: str, prompt: str):
        raise NotImplementedError

    @abstractmethod
    def _func_fail_default_resp(self):
        raise NotImplementedError

    def generate_prompt_with_tmpl_filename(self, prompt_input: Union[str, list], tmpl_filename) -> str:
        """
        same with `generate_prompt`
        Args:
            prompt_input: the input we want to feed in (IF THERE ARE MORE THAN ONE INPUT, THIS CAN BE A LIST.)
            tmpl_filename: prompt template filename
        Returns:
            a str prompt that will be sent to LLM server.
        """
        if isinstance(prompt_input, str):
            prompt_input = [prompt_input]
        prompt_input = [str(i) for i in prompt_input]

        f = open(str(self.prompt_dir.joinpath(tmpl_filename)), "r")
        prompt = f.read()
        f.close()
        for count, i in enumerate(prompt_input):
            prompt = prompt.replace(f"!<INPUT {count}>!", i)
        if "<commentblockmarker>###</commentblockmarker>" in prompt:
            prompt = prompt.split("<commentblockmarker>###</commentblockmarker>")[1]
        return prompt.strip()

    async def _aask(self, prompt: str) -> str:
        return await self.llm.aask(prompt)

    async def _run_gpt35_max_tokens(self, prompt: str, max_tokens: int = 50, retry: int = 3):
        for idx in range(retry):
            try:
                tmp_max_tokens_rsp = getattr(config.llm, "max_token", 1500)
                setattr(config.llm, "max_token", max_tokens)
                self.llm.use_system_prompt = False  # to make it behave like a non-chat completions

                llm_resp = await self._aask(prompt)

                setattr(config.llm, "max_token", tmp_max_tokens_rsp)
                logger.info(f"Action: {self.cls_name} llm _run_gpt35_max_tokens raw resp: {llm_resp}")
                if self._func_validate(llm_resp, prompt):
                    return self._func_cleanup(llm_resp, prompt)
            except Exception as exp:
                logger.warning(f"Action: {self.cls_name} _run_gpt35_max_tokens exp: {exp}")
                time.sleep(5)
        return self.fail_default_resp

    async def _run_gpt35(
        self, prompt: str, example_output: str, special_instruction: str, retry: int = 3
    ) -> Union[bool, Any]:
        """same with `gpt_structure.ChatGPT_safe_generate_response`"""
        prompt = '"""\n' + prompt + '\n"""\n'
        prompt += f"Output the response to the prompt above in json. {special_instruction}\n"
        prompt += "Example output json:\n"
        prompt += '{"output": "' + str(example_output) + '"}'

        for idx in range(retry):
            try:
                llm_resp = await self._aask(prompt)
                logger.info(f"Action: {self.cls_name} llm _run_gpt35 raw resp: {llm_resp}")
                end_idx = llm_resp.strip().rfind("}") + 1
                llm_resp = llm_resp[:end_idx]
                llm_resp = json.loads(llm_resp)["output"]

                if self._func_validate(llm_resp, prompt):
                    return self._func_cleanup(llm_resp, prompt)
            except Exception as exp:
                logger.warning(f"Action: {self.cls_name} _run_gpt35 exp: {exp}")
                time.sleep(5)  # usually avoid `Rate limit`
        return False

    async def _run_gpt35_wo_extra_prompt(self, prompt: str, retry: int = 3) -> str:
        for idx in range(retry):
            try:
                llm_resp = await self._aask(prompt)
                llm_resp = llm_resp.strip()
                logger.info(f"Action: {self.cls_name} llm _run_gpt35_wo_extra_prompt raw resp: {llm_resp}")
                if self._func_validate(llm_resp, prompt):
                    return self._func_cleanup(llm_resp, prompt)
            except Exception as exp:
                logger.warning(f"Action: {self.cls_name} _run_gpt35_wo_extra_prompt exp: {exp}")
                time.sleep(5)  # usually avoid `Rate limit`
        return self.fail_default_resp

    async def run(self, *args, **kwargs):
        """Run action"""
        raise NotImplementedError("The run method should be implemented in a subclass.")



import mimetypes
import os
from pathlib import Path
from typing import Union

import httpx

from metagpt.rag.schema import OmniParsedResult
from metagpt.utils.common import aread_bin


class OmniParseClient:
    """
    OmniParse Server Client
    This client interacts with the OmniParse server to parse different types of media, documents.

    OmniParse API Documentation: https://docs.cognitivelab.in/api

    Attributes:
        ALLOWED_DOCUMENT_EXTENSIONS (set): A set of supported document file extensions.
        ALLOWED_AUDIO_EXTENSIONS (set): A set of supported audio file extensions.
        ALLOWED_VIDEO_EXTENSIONS (set): A set of supported video file extensions.
    """

    ALLOWED_DOCUMENT_EXTENSIONS = {".pdf", ".ppt", ".pptx", ".doc", ".docx"}
    ALLOWED_AUDIO_EXTENSIONS = {".mp3", ".wav", ".aac"}
    ALLOWED_VIDEO_EXTENSIONS = {".mp4", ".mkv", ".avi", ".mov"}

    def __init__(self, api_key: str = None, base_url: str = "http://localhost:8000", max_timeout: int = 120):
        """
        Args:
            api_key: Default None, can be used for authentication later.
            base_url: Base URL for the API.
            max_timeout: Maximum request timeout in seconds.
        """
        self.api_key = api_key
        self.base_url = base_url
        self.max_timeout = max_timeout

        self.parse_media_endpoint = "/parse_media"
        self.parse_website_endpoint = "/parse_website"
        self.parse_document_endpoint = "/parse_document"

    async def _request_parse(
        self,
        endpoint: str,
        method: str = "POST",
        files: dict = None,
        params: dict = None,
        data: dict = None,
        json: dict = None,
        headers: dict = None,
        **kwargs,
    ) -> dict:
        """
        Request OmniParse API to parse a document.

        Args:
            endpoint (str): API endpoint.
            method (str, optional): HTTP method to use. Default is "POST".
            files (dict, optional): Files to include in the request.
            params (dict, optional): Query string parameters.
            data (dict, optional): Form data to include in the request body.
            json (dict, optional): JSON data to include in the request body.
            headers (dict, optional): HTTP headers to include in the request.
            **kwargs: Additional keyword arguments for httpx.AsyncClient.request()

        Returns:
            dict: JSON response data.
        """
        url = f"{self.base_url}{endpoint}"
        method = method.upper()
        headers = headers or {}
        _headers = {"Authorization": f"Bearer {self.api_key}"} if self.api_key else {}
        headers.update(**_headers)
        async with httpx.AsyncClient() as client:
            response = await client.request(
                url=url,
                method=method,
                files=files,
                params=params,
                json=json,
                data=data,
                headers=headers,
                timeout=self.max_timeout,
                **kwargs,
            )
            response.raise_for_status()
            return response.json()

    async def parse_document(self, file_input: Union[str, bytes, Path], bytes_filename: str = None) -> OmniParsedResult:
        """
        Parse document-type data (supports ".pdf", ".ppt", ".pptx", ".doc", ".docx").

        Args:
            file_input: File path or file byte data.
            bytes_filename: Filename for byte data, useful for determining MIME type for the HTTP request.

        Raises:
            ValueError: If the file extension is not allowed.

        Returns:
            OmniParsedResult: The result of the document parsing.
        """
        self.verify_file_ext(file_input, self.ALLOWED_DOCUMENT_EXTENSIONS, bytes_filename)
        file_info = await self.get_file_info(file_input, bytes_filename)
        resp = await self._request_parse(self.parse_document_endpoint, files={"file": file_info})
        data = OmniParsedResult(**resp)
        return data

    async def parse_pdf(self, file_input: Union[str, bytes, Path]) -> OmniParsedResult:
        """
        Parse pdf document.

        Args:
            file_input: File path or file byte data.

        Raises:
            ValueError: If the file extension is not allowed.

        Returns:
            OmniParsedResult: The result of the pdf parsing.
        """
        self.verify_file_ext(file_input, {".pdf"})
        # parse_pdf supports parsing by accepting only the byte data of the file.
        file_info = await self.get_file_info(file_input, only_bytes=True)
        endpoint = f"{self.parse_document_endpoint}/pdf"
        resp = await self._request_parse(endpoint=endpoint, files={"file": file_info})
        data = OmniParsedResult(**resp)
        return data

    async def parse_video(self, file_input: Union[str, bytes, Path], bytes_filename: str = None) -> dict:
        """
        Parse video-type data (supports ".mp4", ".mkv", ".avi", ".mov").

        Args:
            file_input: File path or file byte data.
            bytes_filename: Filename for byte data, useful for determining MIME type for the HTTP request.

        Raises:
            ValueError: If the file extension is not allowed.

        Returns:
            dict: JSON response data.
        """
        self.verify_file_ext(file_input, self.ALLOWED_VIDEO_EXTENSIONS, bytes_filename)
        file_info = await self.get_file_info(file_input, bytes_filename)
        return await self._request_parse(f"{self.parse_media_endpoint}/video", files={"file": file_info})

    async def parse_audio(self, file_input: Union[str, bytes, Path], bytes_filename: str = None) -> dict:
        """
        Parse audio-type data (supports ".mp3", ".wav", ".aac").

        Args:
            file_input: File path or file byte data.
            bytes_filename: Filename for byte data, useful for determining MIME type for the HTTP request.

        Raises:
            ValueError: If the file extension is not allowed.

        Returns:
            dict: JSON response data.
        """
        self.verify_file_ext(file_input, self.ALLOWED_AUDIO_EXTENSIONS, bytes_filename)
        file_info = await self.get_file_info(file_input, bytes_filename)
        return await self._request_parse(f"{self.parse_media_endpoint}/audio", files={"file": file_info})

    @staticmethod
    def verify_file_ext(file_input: Union[str, bytes, Path], allowed_file_extensions: set, bytes_filename: str = None):
        """
        Verify the file extension.

        Args:
            file_input: File path or file byte data.
            allowed_file_extensions: Set of allowed file extensions.
            bytes_filename: Filename to use for verification when `file_input` is byte data.

        Raises:
            ValueError: If the file extension is not allowed.

        Returns:
        """
        verify_file_path = None
        if isinstance(file_input, (str, Path)):
            verify_file_path = str(file_input)
        elif isinstance(file_input, bytes) and bytes_filename:
            verify_file_path = bytes_filename

        if not verify_file_path:
            # Do not verify if only byte data is provided
            return

        file_ext = os.path.splitext(verify_file_path)[1].lower()
        if file_ext not in allowed_file_extensions:
            raise ValueError(f"Not allowed {file_ext} File extension must be one of {allowed_file_extensions}")

    @staticmethod
    async def get_file_info(
        file_input: Union[str, bytes, Path],
        bytes_filename: str = None,
        only_bytes: bool = False,
    ) -> Union[bytes, tuple]:
        """
        Get file information.

        Args:
            file_input: File path or file byte data.
            bytes_filename: Filename to use when uploading byte data, useful for determining MIME type.
            only_bytes: Whether to return only byte data. Default is False, which returns a tuple.

        Raises:
            ValueError: If bytes_filename is not provided when file_input is bytes or if file_input is not a valid type.

        Notes:
            Since `parse_document`,`parse_video`, `parse_audio` supports parsing various file types,
            the MIME type of the file must be specified when uploading.

        Returns: [bytes, tuple]
            Returns bytes if only_bytes is True, otherwise returns a tuple (filename, file_bytes, mime_type).
        """
        if isinstance(file_input, (str, Path)):
            filename = os.path.basename(str(file_input))
            file_bytes = await aread_bin(file_input)

            if only_bytes:
                return file_bytes

            mime_type = mimetypes.guess_type(file_input)[0]
            return filename, file_bytes, mime_type
        elif isinstance(file_input, bytes):
            if only_bytes:
                return file_input
            if not bytes_filename:
                raise ValueError("bytes_filename must be set when passing bytes")

            mime_type = mimetypes.guess_type(bytes_filename)[0]
            return bytes_filename, file_input, mime_type
        else:
            raise ValueError("file_input must be a string (file path) or bytes.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Desc   : parse record to generate learned standard operations in stage=learn & mode=manual,
#           LIKE scripts/document_generation.py

import ast
import re
from pathlib import Path

from metagpt.actions.action import Action
from metagpt.config2 import config
from metagpt.ext.android_assistant.actions.parse_record_an import RECORD_PARSE_NODE
from metagpt.ext.android_assistant.prompts.operation_prompt import (
    long_press_doc_template,
    refine_doc_suffix,
    swipe_doc_template,
    tap_doc_template,
    text_doc_template,
)
from metagpt.ext.android_assistant.utils.schema import (
    ActionOp,
    AndroidActionOutput,
    RecordLogItem,
    RunState,
    SwipeOp,
)
from metagpt.logs import logger
from metagpt.utils.common import encode_image


class ParseRecord(Action):
    name: str = "ParseRecord"
    record_path: Path = ""
    task_desc_path: Path = ""
    screenshot_before_path: Path = ""
    screenshot_after_path: Path = ""

    async def run(self, task_dir: Path, docs_dir: Path):
        doc_count = 0
        self.record_path = Path(task_dir) / "record.txt"
        self.task_desc_path = Path(task_dir) / "task_desc.txt"
        self.screenshot_before_path = Path(task_dir) / "raw_screenshots"
        self.screenshot_after_path = Path(task_dir) / "labeled_screenshots"
        for path in [self.screenshot_before_path, self.screenshot_after_path]:
            path.mkdir(parents=True, exist_ok=True)

        task_desc = self.task_desc_path.read_text()
        extra_config = config.extra

        with open(self.record_path, "r") as record_file:
            record_step_count = len(record_file.readlines()) - 1
            record_file.seek(0)
            for step in range(1, record_step_count + 1):
                img_before_base64 = encode_image(self.screenshot_after_path.joinpath(f"{step}_labeled.png"))
                img_after_base64 = encode_image(self.screenshot_after_path.joinpath(f"{step + 1}_labeled.png"))
                rec = record_file.readline().strip()
                action, resource_id = rec.split(":::")
                action_type = action.split("(")[0]
                # 构建Prompt
                action_param = re.findall(r"\((.*?)\)", action)[0]
                if action_type == ActionOp.TAP.value:
                    prompt_template = tap_doc_template
                    context = prompt_template.format(ui_element=action_param)
                elif action_type == ActionOp.TEXT.value:
                    input_area, input_text = action_param.split(":sep:")
                    prompt_template = text_doc_template
                    context = prompt_template.format(ui_element=input_area)
                elif action_type == ActionOp.LONG_PRESS.value:
                    prompt_template = long_press_doc_template
                    context = prompt_template.format(ui_element=action_param)
                elif action_type == ActionOp.SWIPE.value:
                    swipe_area, swipe_dir = action_param.split(":sep:")
                    if swipe_dir == SwipeOp.UP.value or swipe_dir == SwipeOp.DOWN.value:
                        action_type = ActionOp.VERTICAL_SWIPE.value
                    elif swipe_dir == SwipeOp.LEFT.value or swipe_dir == SwipeOp.RIGHT.value:
                        action_type = ActionOp.HORIZONTAL_SWIPE.value
                    prompt_template = swipe_doc_template
                    context = prompt_template.format(swipe_dir=swipe_dir, ui_element=swipe_area)
                else:
                    break
                context = context.format(task_desc=task_desc)

                doc_name = resource_id + ".txt"
                doc_path = docs_dir.joinpath(doc_name)

                if doc_path.exists():
                    try:
                        doc_content = ast.literal_eval(doc_path.read_text())
                    except Exception as exp:
                        logger.error(f"ast parse doc: {doc_path} failed, exp: {exp}")
                        continue

                    if doc_content[action_type]:
                        if extra_config.get("doc_refine", False):
                            refine_context = refine_doc_suffix.format(old_doc=doc_content[action_type])
                            context += refine_context
                            logger.info(
                                f"Documentation for the element {resource_id} already exists. The doc will be "
                                f"refined based on the latest demo."
                            )
                        else:
                            logger.info(
                                f"Documentation for the element {resource_id} already exists. Turn on DOC_REFINE "
                                f"in the config file if needed."
                            )
                            continue
                else:
                    doc_content = {"tap": "", "text": "", "v_swipe": "", "h_swipe": "", "long_press": ""}

                logger.info(f"Waiting for GPT-4V to generate documentation for the element {resource_id}")
                node = await RECORD_PARSE_NODE.fill(
                    context=context, llm=self.llm, images=[img_before_base64, img_after_base64]
                )
                if "error" in node.content:
                    return AndroidActionOutput(action_state=RunState.FAIL)
                log_path = task_dir.joinpath("log_parse_record.txt")
                prompt = node.compile(context=context, schema="json", mode="auto")
                msg = node.content
                doc_content[action_type] = msg

                with open(log_path, "a") as logfile:
                    log_item = RecordLogItem(
                        step=step,
                        prompt=prompt,
                        image_before=img_before_base64,
                        image_after=img_after_base64,
                        response=node.content,
                    )
                    logfile.write(log_item.model_dump_json() + "\n")
                with open(doc_path, "w") as outfile:
                    outfile.write(str(doc_content))
                doc_count += 1
                logger.info(f"Documentation generated and saved to {doc_path}")

            logger.info(f"Documentation generation phase completed. {doc_count} docs generated.")

        return AndroidActionOutput(action_state=RunState.FINISH)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Desc   :

from metagpt.ext.werewolf.actions.werewolf_actions import Hunt, Impersonate
from metagpt.ext.werewolf.actions.guard_actions import Protect
from metagpt.ext.werewolf.actions.seer_actions import Verify
from metagpt.ext.werewolf.actions.witch_actions import Save, Poison
from metagpt.ext.werewolf.actions.common_actions import Speak, NighttimeWhispers, Reflect
from metagpt.ext.werewolf.actions.experience_operation import AddNewExperiences, RetrieveExperiences
from metagpt.ext.werewolf.actions.moderator_actions import InstructSpeak

ACTIONS = {
    "Speak": Speak,
    "Hunt": Hunt,
    "Protect": Protect,
    "Verify": Verify,
    "Save": Save,
    "Poison": Poison,
    "Impersonate": Impersonate,
}

__all__ = ["NighttimeWhispers", "Reflect", "AddNewExperiences", "RetrieveExperiences", "InstructSpeak"]



from __future__ import annotations

import json
from typing import Literal

from pydantic import Field, model_validator

from metagpt.actions.di.ask_review import ReviewConst
from metagpt.actions.di.execute_nb_code import ExecuteNbCode
from metagpt.actions.di.write_analysis_code import CheckData, WriteAnalysisCode
from metagpt.logs import logger
from metagpt.prompts.di.write_analysis_code import DATA_INFO
from metagpt.roles import Role
from metagpt.schema import Message, Task, TaskResult
from metagpt.strategy.task_type import TaskType
from metagpt.tools.tool_recommend import BM25ToolRecommender, ToolRecommender
from metagpt.utils.common import CodeParser

REACT_THINK_PROMPT = """
# User Requirement
{user_requirement}
# Context
{context}

Output a json following the format:
```json
{{
    "thoughts": str = "Thoughts on current situation, reflect on how you should proceed to fulfill the user requirement",
    "state": bool = "Decide whether you need to take more actions to complete the user requirement. Return true if you think so. Return false if you think the requirement has been completely fulfilled."
}}
```
"""


class DataInterpreter(Role):
    name: str = "David"
    profile: str = "DataInterpreter"
    auto_run: bool = True
    use_plan: bool = True
    use_reflection: bool = False
    execute_code: ExecuteNbCode = Field(default_factory=ExecuteNbCode, exclude=True)
    tools: list[str] = []  # Use special symbol ["<all>"] to indicate use of all registered tools
    tool_recommender: ToolRecommender = None
    react_mode: Literal["plan_and_act", "react"] = "plan_and_act"
    max_react_loop: int = 10  # used for react mode

    @model_validator(mode="after")
    def set_plan_and_tool(self) -> "Interpreter":
        self._set_react_mode(react_mode=self.react_mode, max_react_loop=self.max_react_loop, auto_run=self.auto_run)
        self.use_plan = (
            self.react_mode == "plan_and_act"
        )  # create a flag for convenience, overwrite any passed-in value
        if self.tools and not self.tool_recommender:
            self.tool_recommender = BM25ToolRecommender(tools=self.tools)
        self.set_actions([WriteAnalysisCode])
        self._set_state(0)
        return self

    @property
    def working_memory(self):
        return self.rc.working_memory

    async def _think(self) -> bool:
        """Useful in 'react' mode. Use LLM to decide whether and what to do next."""
        user_requirement = self.get_memories()[0].content
        context = self.working_memory.get()

        if not context:
            # just started the run, we need action certainly
            self.working_memory.add(self.get_memories()[0])  # add user requirement to working memory
            self._set_state(0)
            return True

        prompt = REACT_THINK_PROMPT.format(user_requirement=user_requirement, context=context)
        rsp = await self.llm.aask(prompt)
        rsp_dict = json.loads(CodeParser.parse_code(block=None, text=rsp))
        self.working_memory.add(Message(content=rsp_dict["thoughts"], role="assistant"))
        need_action = rsp_dict["state"]
        self._set_state(0) if need_action else self._set_state(-1)

        return need_action

    async def _act(self) -> Message:
        """Useful in 'react' mode. Return a Message conforming to Role._act interface."""
        code, _, _ = await self._write_and_exec_code()
        return Message(content=code, role="assistant", cause_by=WriteAnalysisCode)

    async def _plan_and_act(self) -> Message:
        try:
            rsp = await super()._plan_and_act()
            await self.execute_code.terminate()
            return rsp
        except Exception as e:
            await self.execute_code.terminate()
            raise e

    async def _act_on_task(self, current_task: Task) -> TaskResult:
        """Useful in 'plan_and_act' mode. Wrap the output in a TaskResult for review and confirmation."""
        code, result, is_success = await self._write_and_exec_code()
        task_result = TaskResult(code=code, result=result, is_success=is_success)
        return task_result

    async def _write_and_exec_code(self, max_retry: int = 3):
        counter = 0
        success = False

        # plan info
        plan_status = self.planner.get_plan_status() if self.use_plan else ""

        # tool info
        if self.tool_recommender:
            context = (
                self.working_memory.get()[-1].content if self.working_memory.get() else ""
            )  # thoughts from _think stage in 'react' mode
            plan = self.planner.plan if self.use_plan else None
            tool_info = await self.tool_recommender.get_recommended_tool_info(context=context, plan=plan)
        else:
            tool_info = ""

        # data info
        await self._check_data()

        while not success and counter < max_retry:
            ### write code ###
            code, cause_by = await self._write_code(counter, plan_status, tool_info)

            self.working_memory.add(Message(content=code, role="assistant", cause_by=cause_by))

            ### execute code ###
            result, success = await self.execute_code.run(code)
            print(result)

            self.working_memory.add(Message(content=result, role="user", cause_by=ExecuteNbCode))

            ### process execution result ###
            counter += 1

            if not success and counter >= max_retry:
                logger.info("coding failed!")
                review, _ = await self.planner.ask_review(auto_run=False, trigger=ReviewConst.CODE_REVIEW_TRIGGER)
                if ReviewConst.CHANGE_WORDS[0] in review:
                    counter = 0  # redo the task again with help of human suggestions

        return code, result, success

    async def _write_code(
        self,
        counter: int,
        plan_status: str = "",
        tool_info: str = "",
    ):
        todo = self.rc.todo  # todo is WriteAnalysisCode
        logger.info(f"ready to {todo.name}")
        use_reflection = counter > 0 and self.use_reflection  # only use reflection after the first trial

        user_requirement = self.get_memories()[0].content

        code = await todo.run(
            user_requirement=user_requirement,
            plan_status=plan_status,
            tool_info=tool_info,
            working_memory=self.working_memory.get(),
            use_reflection=use_reflection,
        )

        return code, todo

    async def _check_data(self):
        if (
            not self.use_plan
            or not self.planner.plan.get_finished_tasks()
            or self.planner.plan.current_task.task_type
            not in [
                TaskType.DATA_PREPROCESS.type_name,
                TaskType.FEATURE_ENGINEERING.type_name,
                TaskType.MODEL_TRAIN.type_name,
            ]
        ):
            return
        logger.info("Check updated data")
        code = await CheckData().run(self.planner.plan)
        if not code.strip():
            return
        result, success = await self.execute_code.run(code)
        if success:
            print(result)
            data_info = DATA_INFO.format(info=result)
            self.working_memory.add(Message(content=data_info, role="user", cause_by=CheckData))



"""Base Ranker."""

from abc import abstractmethod
from typing import Optional

from llama_index.core.postprocessor.types import BaseNodePostprocessor
from llama_index.core.schema import NodeWithScore, QueryBundle


class RAGRanker(BaseNodePostprocessor):
    """inherit from llama_index"""

    @abstractmethod
    def _postprocess_nodes(
        self,
        nodes: list[NodeWithScore],
        query_bundle: Optional[QueryBundle] = None,
    ) -> list[NodeWithScore]:
        """postprocess nodes."""



from typing import Any, Optional

from metagpt.actions.add_requirement import UserRequirement
from metagpt.context import Context
from metagpt.environment.werewolf.werewolf_env import WerewolfEnv
from metagpt.ext.werewolf.schema import WwMessage
from metagpt.team import Team


class WerewolfGame(Team):
    """Use the "software company paradigm" to hold a werewolf game"""

    env: Optional[WerewolfEnv] = None

    def __init__(self, context: Context = None, **data: Any):
        super(Team, self).__init__(**data)
        ctx = context or Context()
        if not self.env:
            self.env = WerewolfEnv(context=ctx)
        else:
            self.env.context = ctx  # The `env` object is allocated by deserialization

    def run_project(self, idea):
        """Run a project from user instruction."""
        self.idea = idea
        self.env.publish_message(
            WwMessage(role="User", content=idea, cause_by=UserRequirement, restricted_to={"Moderator"})
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import annotations

import json
import warnings
from typing import Optional

import aiohttp
from pydantic import BaseModel, ConfigDict, model_validator


class BingAPIWrapper(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    api_key: str
    bing_url: str = "https://api.bing.microsoft.com/v7.0/search"
    aiosession: Optional[aiohttp.ClientSession] = None
    proxy: Optional[str] = None

    @model_validator(mode="before")
    @classmethod
    def validate_api_key(cls, values: dict) -> dict:
        if "api_key" in values:
            values.setdefault("api_key", values["api_key"])
            warnings.warn("`api_key` is deprecated, use `api_key` instead", DeprecationWarning, stacklevel=2)
        return values

    @property
    def header(self):
        return {"Ocp-Apim-Subscription-Key": self.api_key}

    async def run(
        self,
        query: str,
        max_results: int = 8,
        as_string: bool = True,
        focus: list[str] | None = None,
    ) -> str | list[dict]:
        """Return the results of a Google search using the official Bing API.

        Args:
            query: The search query.
            max_results: The number of results to return.
            as_string: A boolean flag to determine the return type of the results. If True, the function will
                return a formatted string with the search results. If False, it will return a list of dictionaries
                containing detailed information about each search result.
            focus: Specific information to be focused on from each search result.

        Returns:
            The results of the search.
        """
        params = {
            "q": query,
            "count": max_results,
            "textFormat": "HTML",
        }
        result = await self.results(params)
        search_results = result["webPages"]["value"]
        focus = focus or ["snippet", "link", "title"]
        for item_dict in search_results:
            item_dict["link"] = item_dict["url"]
            item_dict["title"] = item_dict["name"]
        details = [{i: j for i, j in item_dict.items() if i in focus} for item_dict in search_results]
        if as_string:
            return safe_results(details)
        return details

    async def results(self, params: dict) -> dict:
        """Use aiohttp to run query and return the results async."""

        if not self.aiosession:
            async with aiohttp.ClientSession() as session:
                async with session.get(self.bing_url, params=params, headers=self.header, proxy=self.proxy) as response:
                    response.raise_for_status()
                    res = await response.json()
        else:
            async with self.aiosession.get(
                self.bing_url, params=params, headers=self.header, proxy=self.proxy
            ) as response:
                response.raise_for_status()
                res = await response.json()

        return res


def safe_results(results: str | list) -> str:
    """Return the results of a bing search in a safe format.

    Args:
        results: The search results.

    Returns:
        The results of the search.
    """
    if isinstance(results, list):
        safe_message = json.dumps([result for result in results])
    else:
        safe_message = results.encode("utf-8", "ignore").decode("utf-8")
    return safe_message


if __name__ == "__main__":
    import fire

    fire.Fire(BingAPIWrapper().run)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Desc   : task_decomp

import datetime

from metagpt.ext.stanford_town.actions.st_action import STAction
from metagpt.logs import logger


class TaskDecomp(STAction):
    name: str = "TaskDecomp"

    def _func_cleanup(self, llm_resp: str, prompt: str) -> list:
        # TODO SOMETHING HERE sometimes fails... See screenshot
        temp = [i.strip() for i in llm_resp.split("\n")]
        _cr = []
        cr = []
        for count, i in enumerate(temp):
            if count != 0:
                _cr += [" ".join([j.strip() for j in i.split(" ")][3:])]
            else:
                _cr += [i]
        for count, i in enumerate(_cr):
            k = [j.strip() for j in i.split("(duration in minutes:")]
            task = k[0]
            if task[-1] == ".":
                task = task[:-1]
            duration = int(k[1].split(",")[0].strip())
            cr += [[task, duration]]

        total_expected_min = int(prompt.split("(total duration in minutes")[-1].split("):")[0].strip())

        # TODO -- now, you need to make sure that this is the same as the sum of
        #         the current action sequence.
        curr_min_slot = [
            ["dummy", -1],
        ]  # (task_name, task_index)
        for count, i in enumerate(cr):
            i_task = i[0]
            i_duration = i[1]

            i_duration -= i_duration % 5
            if i_duration > 0:
                for j in range(i_duration):
                    curr_min_slot += [(i_task, count)]
        curr_min_slot = curr_min_slot[1:]

        if len(curr_min_slot) > total_expected_min:
            last_task = curr_min_slot[60]
            for i in range(1, 6):
                curr_min_slot[-1 * i] = last_task
        elif len(curr_min_slot) < total_expected_min:
            last_task = curr_min_slot[-1]
            for i in range(total_expected_min - len(curr_min_slot)):
                curr_min_slot += [last_task]

        cr_ret = [
            ["dummy", -1],
        ]
        for task, task_index in curr_min_slot:
            if task != cr_ret[-1][0]:
                cr_ret += [[task, 1]]
            else:
                cr_ret[-1][1] += 1
        cr = cr_ret[1:]

        return cr

    def _func_validate(self, llm_resp: str, prompt: str) -> bool:
        # TODO -- this sometimes generates error
        try:
            self._func_cleanup(llm_resp, prompt)
        except Exception:
            return False
        return True

    def _func_fail_default_resp(self) -> int:
        fs = [["asleep", 0]]
        return fs

    async def run(self, role: "STRole", task_desc: int, truncated_act_dur: int, *args, **kwargs):
        def create_prompt_input(role, task, duration):
            """
            Today is Saturday June 25. From 00:00 ~ 06:00am, Maeve is
            planning on sleeping, 06:00 ~ 07:00am, Maeve is
            planning on waking up and doing her morning routine,
            and from 07:00am ~08:00am, Maeve is planning on having breakfast.
            """

            curr_f_org_index = role.scratch.get_f_daily_schedule_hourly_org_index()
            all_indices = []
            # if curr_f_org_index > 0:
            #   all_indices += [curr_f_org_index-1]
            all_indices += [curr_f_org_index]
            if curr_f_org_index + 1 <= len(role.scratch.f_daily_schedule_hourly_org):
                all_indices += [curr_f_org_index + 1]
            if curr_f_org_index + 2 <= len(role.scratch.f_daily_schedule_hourly_org):
                all_indices += [curr_f_org_index + 2]

            curr_time_range = ""

            logger.debug("DEBUG")
            logger.debug(role.scratch.f_daily_schedule_hourly_org)
            logger.debug(all_indices)

            summ_str = f'Today is {role.scratch.curr_time.strftime("%B %d, %Y")}. '
            summ_str += "From "
            for index in all_indices:
                logger.debug(f"index {index}")
                if index < len(role.scratch.f_daily_schedule_hourly_org):
                    start_min = 0
                    for i in range(index):
                        start_min += role.scratch.f_daily_schedule_hourly_org[i][1]
                        end_min = start_min + role.scratch.f_daily_schedule_hourly_org[index][1]
                        start_time = datetime.datetime.strptime("00:00:00", "%H:%M:%S") + datetime.timedelta(
                            minutes=start_min
                        )
                        end_time = datetime.datetime.strptime("00:00:00", "%H:%M:%S") + datetime.timedelta(
                            minutes=end_min
                        )
                        start_time_str = start_time.strftime("%H:%M%p")
                        end_time_str = end_time.strftime("%H:%M%p")
                        summ_str += (
                            f"{start_time_str} ~ {end_time_str}, {role.name} is planning "
                            f"on {role.scratch.f_daily_schedule_hourly_org[index][0]}, "
                        )
                        if curr_f_org_index + 1 == index:
                            curr_time_range = f"{start_time_str} ~ {end_time_str}"
            summ_str = summ_str[:-2] + "."

            prompt_input = []
            prompt_input += [role.scratch.get_str_iss()]
            prompt_input += [summ_str]
            # prompt_input += [role.scratch.get_str_curr_date_str()]
            prompt_input += [role.scratch.get_str_firstname()]
            prompt_input += [role.scratch.get_str_firstname()]
            prompt_input += [task]
            prompt_input += [curr_time_range]
            prompt_input += [duration]
            prompt_input += [role.scratch.get_str_firstname()]
            return prompt_input

        prompt_input = create_prompt_input(role, task_desc, truncated_act_dur)
        prompt = self.generate_prompt_with_tmpl_filename(prompt_input, "task_decomp_v3.txt")
        self.fail_default_resp = self._func_fail_default_resp()
        output = await self._run_gpt35_max_tokens(prompt, max_tokens=1000)
        logger.info(f"Role: {role.name} {self.cls_name} output: {output}")

        fin_output = []
        time_sum = 0
        for i_task, i_duration in output:
            time_sum += i_duration
            # HM?????????
            # if time_sum < duration:
            if time_sum <= truncated_act_dur:
                fin_output += [[i_task, i_duration]]
            else:
                break
        ftime_sum = 0
        for fi_task, fi_duration in fin_output:
            ftime_sum += fi_duration

        fin_output[-1][1] += truncated_act_dur - ftime_sum
        output = fin_output

        task_decomp = output
        ret = []
        for decomp_task, duration in task_decomp:
            ret += [[f"{task_desc} ({decomp_task})", duration]]
        output = ret
        logger.info(f"Role: {role.name} Action: {self.cls_name} output: {output}")
        return output



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Desc   :

from enum import Enum

from pydantic import BaseModel, Field, field_validator


class ActionOp(Enum):
    TAP = "tap"
    LONG_PRESS = "long_press"
    TEXT = "text"
    SWIPE = "swipe"
    VERTICAL_SWIPE = "v_swipe"
    HORIZONTAL_SWIPE = "h_swipe"
    GRID = "grid"
    STOP = "stop"


class SwipeOp(Enum):
    UP = "up"
    DOWN = "down"
    LEFT = "left"
    RIGHT = "right"


class Decision(Enum):
    BACK = "BACK"
    INEFFECTIVE = "INEFFECTIVE"
    CONTINUE = "CONTINUE"
    SUCCESS = "SUCCESS"

    @classmethod
    def values(cls):
        return [item.value for item in cls]


class AndroidElement(BaseModel):
    """UI Element"""

    uid: str = Field(default="")
    bbox: tuple[tuple[int, int], tuple[int, int]] = Field(default={})
    attrib: str = Field(default="")


class OpLogItem(BaseModel):
    """log content for self-learn or task act"""

    step: int = Field(default=0)
    prompt: str = Field(default="")
    image: str = Field(default="")
    response: str = Field(default="")


class ReflectLogItem(BaseModel):
    """log content for self-learn-reflect"""

    step: int = Field(default=0)
    prompt: str = Field(default="")
    image_before: str = Field(default="")
    image_after: str = Field(default="")
    response: str = Field(default="")


class RecordLogItem(BaseModel):
    """log content for record parse, same as ReflectLogItem"""

    step: int = Field(default=0)
    prompt: str = Field(default="")
    image_before: str = Field(default="")
    image_after: str = Field(default="")
    response: str = Field(default="")


class DocContent(BaseModel):
    tap: str = Field(default="")
    text: str = Field(default="")
    v_swipe: str = Field(default="")
    h_swipe: str = Field(default="")
    long_press: str = Field(default="")


# start =================== define different Action Op and its params =============
class RunState(Enum):
    """run state"""

    SUCCESS = "success"
    FINISH = "finish"
    FAIL = "fail"


class BaseOpParam(BaseModel):
    act_name: str = Field(default="", validate_default=True)
    last_act: str = Field(default="None")
    param_state: RunState = Field(default=RunState.SUCCESS, description="return state when extract params")


class TapOpParam(BaseOpParam):
    area: int = Field(default=-1)


class TextOpParam(BaseOpParam):
    input_str: str = Field(default="")


class LongPressOpParam(BaseOpParam):
    area: int = Field(default=-1)


# Modify This SwipeOp to SwipeOpParam, Need better name
class SwipeOpParam(BaseOpParam):
    area: int = Field(default=-1)
    swipe_orient: str = Field(default="up")
    dist: str = Field(default="")


class GridOpParam(BaseOpParam):
    act_name: str = Field(default="")


class BaseGridOpParam(BaseOpParam):
    @field_validator("act_name", mode="before")
    @classmethod
    def check_act_name(cls, act_name: str) -> str:
        return f"{act_name}_grid"


class TapGridOpParam(BaseGridOpParam):
    area: int = Field(default=-1)
    subarea: str = Field(default="")


class LongPressGridOpParam(BaseGridOpParam):
    area: int = Field(default=-1)
    subarea: str = Field(default="")


class SwipeGridOpParam(BaseGridOpParam):
    start_area: int = Field(default=-1)
    start_subarea: str = Field(default="")
    end_area: int = Field(default=-1)
    end_subarea: str = Field(default="")


# end =================== define different Action Op and its params =============


class ReflectOp(BaseModel):
    decision: str = ""
    thought: str = ""
    documentation: str = ""
    param_state: RunState = RunState.SUCCESS


class AndroidActionOutput(BaseModel):
    data: dict = Field(default=dict())
    action_state: RunState = Field(default=RunState.SUCCESS)



"""RAG factories"""

from metagpt.rag.factories.retriever import get_retriever
from metagpt.rag.factories.ranker import get_rankers
from metagpt.rag.factories.embedding import get_rag_embedding
from metagpt.rag.factories.index import get_index
from metagpt.rag.factories.llm import get_rag_llm

__all__ = ["get_retriever", "get_rankers", "get_rag_embedding", "get_index", "get_rag_llm"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Desc   :

from metagpt.ext.werewolf.roles.base_player import BasePlayer
from metagpt.ext.werewolf.roles.guard import Guard
from metagpt.ext.werewolf.roles.seer import Seer
from metagpt.ext.werewolf.roles.villager import Villager
from metagpt.ext.werewolf.roles.werewolf import Werewolf
from metagpt.ext.werewolf.roles.witch import Witch
from metagpt.ext.werewolf.roles.moderator import Moderator

__all__ = ["BasePlayer", "Guard", "Moderator", "Seer", "Villager", "Witch", "Werewolf"]



"""RAG Embedding Factory."""
from __future__ import annotations

from typing import Any

from llama_index.core.embeddings import BaseEmbedding
from llama_index.embeddings.azure_openai import AzureOpenAIEmbedding
from llama_index.embeddings.gemini import GeminiEmbedding
from llama_index.embeddings.ollama import OllamaEmbedding
from llama_index.embeddings.openai import OpenAIEmbedding

from metagpt.config2 import config
from metagpt.configs.embedding_config import EmbeddingType
from metagpt.configs.llm_config import LLMType
from metagpt.rag.factories.base import GenericFactory


class RAGEmbeddingFactory(GenericFactory):
    """Create LlamaIndex Embedding with MetaGPT's embedding config."""

    def __init__(self):
        creators = {
            EmbeddingType.OPENAI: self._create_openai,
            EmbeddingType.AZURE: self._create_azure,
            EmbeddingType.GEMINI: self._create_gemini,
            EmbeddingType.OLLAMA: self._create_ollama,
            # For backward compatibility
            LLMType.OPENAI: self._create_openai,
            LLMType.AZURE: self._create_azure,
        }
        super().__init__(creators)

    def get_rag_embedding(self, key: EmbeddingType = None) -> BaseEmbedding:
        """Key is EmbeddingType."""
        return super().get_instance(key or self._resolve_embedding_type())

    def _resolve_embedding_type(self) -> EmbeddingType | LLMType:
        """Resolves the embedding type.

        If the embedding type is not specified, for backward compatibility, it checks if the LLM API type is either OPENAI or AZURE.
        Raise TypeError if embedding type not found.
        """
        if config.embedding.api_type:
            return config.embedding.api_type

        if config.llm.api_type in [LLMType.OPENAI, LLMType.AZURE]:
            return config.llm.api_type

        raise TypeError("To use RAG, please set your embedding in config2.yaml.")

    def _create_openai(self) -> OpenAIEmbedding:
        params = dict(
            api_key=config.embedding.api_key or config.llm.api_key,
            api_base=config.embedding.base_url or config.llm.base_url,
        )

        self._try_set_model_and_batch_size(params)

        return OpenAIEmbedding(**params)

    def _create_azure(self) -> AzureOpenAIEmbedding:
        params = dict(
            api_key=config.embedding.api_key or config.llm.api_key,
            azure_endpoint=config.embedding.base_url or config.llm.base_url,
            api_version=config.embedding.api_version or config.llm.api_version,
        )

        self._try_set_model_and_batch_size(params)

        return AzureOpenAIEmbedding(**params)

    def _create_gemini(self) -> GeminiEmbedding:
        params = dict(
            api_key=config.embedding.api_key,
            api_base=config.embedding.base_url,
        )

        self._try_set_model_and_batch_size(params)

        return GeminiEmbedding(**params)

    def _create_ollama(self) -> OllamaEmbedding:
        params = dict(
            base_url=config.embedding.base_url,
        )

        self._try_set_model_and_batch_size(params)

        return OllamaEmbedding(**params)

    def _try_set_model_and_batch_size(self, params: dict):
        """Set the model_name and embed_batch_size only when they are specified."""
        if config.embedding.model:
            params["model_name"] = config.embedding.model

        if config.embedding.embed_batch_size:
            params["embed_batch_size"] = config.embedding.embed_batch_size

    def _raise_for_key(self, key: Any):
        raise ValueError(f"The embedding type is currently not supported: `{type(key)}`, {key}")


get_rag_embedding = RAGEmbeddingFactory().get_rag_embedding



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from anthropic import AsyncAnthropic
from anthropic.types import Message, Usage

from metagpt.configs.llm_config import LLMConfig, LLMType
from metagpt.const import USE_CONFIG_TIMEOUT
from metagpt.logs import log_llm_stream
from metagpt.provider.base_llm import BaseLLM
from metagpt.provider.llm_provider_registry import register_provider


@register_provider([LLMType.ANTHROPIC, LLMType.CLAUDE])
class AnthropicLLM(BaseLLM):
    def __init__(self, config: LLMConfig):
        self.config = config
        self.__init_anthropic()

    def __init_anthropic(self):
        self.model = self.config.model
        self.aclient: AsyncAnthropic = AsyncAnthropic(api_key=self.config.api_key, base_url=self.config.base_url)

    def _const_kwargs(self, messages: list[dict], stream: bool = False) -> dict:
        kwargs = {
            "model": self.model,
            "messages": messages,
            "max_tokens": self.config.max_token,
            "stream": stream,
        }
        if self.use_system_prompt:
            # if the model support system prompt, extract and pass it
            if messages[0]["role"] == "system":
                kwargs["messages"] = messages[1:]
                kwargs["system"] = messages[0]["content"]  # set system prompt here
        return kwargs

    def _update_costs(self, usage: Usage, model: str = None, local_calc_usage: bool = True):
        usage = {"prompt_tokens": usage.input_tokens, "completion_tokens": usage.output_tokens}
        super()._update_costs(usage, model)

    def get_choice_text(self, resp: Message) -> str:
        return resp.content[0].text

    async def _achat_completion(self, messages: list[dict], timeout: int = USE_CONFIG_TIMEOUT) -> Message:
        resp: Message = await self.aclient.messages.create(**self._const_kwargs(messages))
        self._update_costs(resp.usage, self.model)
        return resp

    async def acompletion(self, messages: list[dict], timeout: int = USE_CONFIG_TIMEOUT) -> Message:
        return await self._achat_completion(messages, timeout=self.get_timeout(timeout))

    async def _achat_completion_stream(self, messages: list[dict], timeout: int = USE_CONFIG_TIMEOUT) -> str:
        stream = await self.aclient.messages.create(**self._const_kwargs(messages, stream=True))
        collected_content = []
        usage = Usage(input_tokens=0, output_tokens=0)
        async for event in stream:
            event_type = event.type
            if event_type == "message_start":
                usage.input_tokens = event.message.usage.input_tokens
                usage.output_tokens = event.message.usage.output_tokens
            elif event_type == "content_block_delta":
                content = event.delta.text
                log_llm_stream(content)
                collected_content.append(content)
            elif event_type == "message_delta":
                usage.output_tokens = event.usage.output_tokens  # update final output_tokens

        log_llm_stream("\n")
        self._update_costs(usage)
        full_content = "".join(collected_content)
        return full_content



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Desc   : conversation between two agents

from typing import Tuple

from metagpt.ext.stanford_town.actions.agent_chat_sum_rel import AgentChatSumRel
from metagpt.ext.stanford_town.actions.gen_iter_chat_utt import GenIterChatUTT
from metagpt.ext.stanford_town.memory.retrieve import new_agent_retrieve
from metagpt.logs import logger


async def agent_conversation(init_role: "STRole", target_role: "STRole", conv_rounds: int = 8) -> list[list[str]]:
    curr_chat = []
    logger.info(f"Role: {init_role.name} starts a conversation with Role: {target_role.name}")

    for idx in range(conv_rounds):
        logger.info(f"Conv round: {idx} between {init_role.name} and {target_role.name}")
        scratch = init_role.rc.scratch
        target_scratch = target_role.rc.scratch

        focal_points = [f"{target_scratch.name}"]
        retrieved = new_agent_retrieve(init_role, focal_points, 50)
        relationship = await generate_summarize_agent_relationship(init_role, target_role, retrieved)
        logger.info(f"The relationship between {init_role.name} and {target_role.name}: {relationship}")
        last_chat = ""
        for i in curr_chat[-4:]:
            last_chat += ": ".join(i) + "\n"
        if last_chat:
            focal_points = [f"{relationship}", f"{target_scratch.name} is {target_scratch.act_description}", last_chat]
        else:
            focal_points = [f"{relationship}", f"{target_scratch.name} is {target_scratch.act_description}"]
        retrieved = new_agent_retrieve(init_role, focal_points, 15)
        utt, end = await generate_one_utterance(init_role, target_role, retrieved, curr_chat)

        curr_chat += [[scratch.name, utt]]
        if end:
            break

        focal_points = [f"{scratch.name}"]
        retrieved = new_agent_retrieve(target_role, focal_points, 50)
        relationship = await generate_summarize_agent_relationship(target_role, init_role, retrieved)
        logger.info(f"The relationship between {target_role.name} and {init_role.name}: {relationship}")
        last_chat = ""
        for i in curr_chat[-4:]:
            last_chat += ": ".join(i) + "\n"
        if last_chat:
            focal_points = [f"{relationship}", f"{scratch.name} is {scratch.act_description}", last_chat]
        else:
            focal_points = [f"{relationship}", f"{scratch.name} is {scratch.act_description}"]
        retrieved = new_agent_retrieve(target_role, focal_points, 15)
        utt, end = await generate_one_utterance(target_role, init_role, retrieved, curr_chat)

        curr_chat += [[target_scratch.name, utt]]
        if end:
            break

    logger.warning(f"Conversations between {target_role.name} and {init_role.name}:")
    for row in curr_chat:
        logger.info(row)

    return curr_chat


async def generate_summarize_agent_relationship(init_role: "STRole", target_role: "STRole", retrieved: dict) -> str:
    all_embedding_keys = list()
    for key, val in retrieved.items():
        for i in val:
            all_embedding_keys += [i.embedding_key]
    all_embedding_key_str = ""
    for i in all_embedding_keys:
        all_embedding_key_str += f"{i}\n"

    summarized_relationship = await AgentChatSumRel().run(init_role, target_role, all_embedding_key_str)
    return summarized_relationship


async def generate_one_utterance(init_role, target_role, retrieved: dict, curr_chat: list) -> Tuple[str, str]:
    # Chat version optimized for speed via batch generation
    scratch = init_role.rc.scratch
    target_scratch = target_role.rc.scratch
    curr_context = (
        f"{scratch.name} "
        + f"was {scratch.act_description} "
        + f"when {scratch.name} "
        + f"saw {target_scratch.name} "
        + f"in the middle of {target_scratch.act_description}.\n"
    )
    curr_context += f"{scratch.name} " + "is initiating a conversation with " + f"{target_scratch.name}."

    x = await GenIterChatUTT().run(init_role, target_role, retrieved, curr_context, curr_chat)

    return x["utterance"], x["end"]



import asyncio
import threading
from typing import Any


def run_coroutine_in_new_loop(coroutine) -> Any:
    """Runs a coroutine in a new, separate event loop on a different thread.

    This function is useful when try to execute an async function within a sync function, but encounter the error `RuntimeError: This event loop is already running`.
    """
    new_loop = asyncio.new_event_loop()
    t = threading.Thread(target=lambda: new_loop.run_forever())
    t.start()

    future = asyncio.run_coroutine_threadsafe(coroutine, new_loop)

    try:
        return future.result()
    finally:
        new_loop.call_soon_threadsafe(new_loop.stop)
        t.join()
        new_loop.close()


class NestAsyncio:
    """Make asyncio event loop reentrant."""

    is_applied = False

    @classmethod
    def apply_once(cls):
        """Ensures `nest_asyncio.apply()` is called only once."""
        if not cls.is_applied:
            import nest_asyncio

            nest_asyncio.apply()
            cls.is_applied = True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Desc   : the implement of memory storage
"""
import shutil
from pathlib import Path

from llama_index.core.embeddings import BaseEmbedding

from metagpt.const import DATA_PATH, MEM_TTL
from metagpt.logs import logger
from metagpt.rag.engines.simple import SimpleEngine
from metagpt.rag.schema import FAISSIndexConfig, FAISSRetrieverConfig
from metagpt.schema import Message
from metagpt.utils.embedding import get_embedding


class MemoryStorage(object):
    """
    The memory storage with Faiss as ANN search engine
    """

    def __init__(self, mem_ttl: int = MEM_TTL, embedding: BaseEmbedding = None):
        self.role_id: str = None
        self.role_mem_path: str = None
        self.mem_ttl: int = mem_ttl  # later use
        self.threshold: float = 0.1  # experience value. TODO The threshold to filter similar memories
        self._initialized: bool = False
        self.embedding = embedding or get_embedding()

        self.faiss_engine = None

    @property
    def is_initialized(self) -> bool:
        return self._initialized

    def recover_memory(self, role_id: str) -> list[Message]:
        self.role_id = role_id
        self.role_mem_path = Path(DATA_PATH / f"role_mem/{self.role_id}/")
        self.role_mem_path.mkdir(parents=True, exist_ok=True)
        self.cache_dir = self.role_mem_path

        if self.role_mem_path.joinpath("default__vector_store.json").exists():
            self.faiss_engine = SimpleEngine.from_index(
                index_config=FAISSIndexConfig(persist_path=self.cache_dir),
                retriever_configs=[FAISSRetrieverConfig()],
                embed_model=self.embedding,
            )
        else:
            self.faiss_engine = SimpleEngine.from_objs(
                objs=[], retriever_configs=[FAISSRetrieverConfig()], embed_model=self.embedding
            )
        self._initialized = True

    def add(self, message: Message) -> bool:
        """add message into memory storage"""
        self.faiss_engine.add_objs([message])
        logger.info(f"Role {self.role_id}'s memory_storage add a message")

    async def search_similar(self, message: Message, k=4) -> list[Message]:
        """search for similar messages"""
        # filter the result which score is smaller than the threshold
        filtered_resp = []
        resp = await self.faiss_engine.aretrieve(message.content)
        for item in resp:
            if item.score < self.threshold:
                filtered_resp.append(item.metadata.get("obj"))
        return filtered_resp

    def clean(self):
        shutil.rmtree(self.cache_dir, ignore_errors=True)
        self._initialized = False

    def persist(self):
        if self.faiss_engine:
            self.faiss_engine.retriever._index.storage_context.persist(self.cache_dir)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Desc   : Retrieve函数实现

import datetime

from numpy import dot
from numpy.linalg import norm

from metagpt.ext.stanford_town.memory.agent_memory import BasicMemory
from metagpt.ext.stanford_town.utils.utils import get_embedding


def agent_retrieve(
    agent_memory,
    curr_time: datetime.datetime,
    memory_forget: float,
    query: str,
    nodes: list[BasicMemory],
    topk: int = 4,
) -> list[BasicMemory]:
    """
    Retrieve需要集合Role使用,原因在于Role才具有AgentMemory,scratch
    逻辑:Role调用该函数,self.rc.AgentMemory,self.rc.scratch.curr_time,self.rc.scratch.memory_forget
    输入希望查询的内容与希望回顾的条数,返回TopK条高分记忆，即List[BasicMemory]

    Score_lists示例
    {
        "memory": memories[i],             BasicMemory类
        "importance": memories[i].poignancy
        "recency": 衰减因子计算结果
        "relevance": 搜索结果
    }
    """
    memories = nodes
    agent_memory_embedding = agent_memory.embeddings
    memories = sorted(memories, key=lambda memory_node: memory_node.last_accessed, reverse=True)

    score_list = []
    score_list = extract_importance(memories, score_list)
    score_list = extract_recency(curr_time, memory_forget, score_list)
    score_list = extract_relevance(agent_memory_embedding, query, score_list)
    score_list = normalize_score_floats(score_list, 0, 1)

    total_dict = {}
    gw = [1, 1, 1]  # 三个因素的权重,重要性,近因性,相关性,
    for i in range(len(score_list)):
        total_score = (
            score_list[i]["importance"] * gw[0] + score_list[i]["recency"] * gw[1] + score_list[i]["relevance"] * gw[2]
        )
        total_dict[score_list[i]["memory"].memory_id] = total_score

    result = top_highest_x_values(total_dict, topk)

    return result  # 返回的是一个BasicMemory列表


def new_agent_retrieve(role, focus_points: list, n_count=30) -> dict:
    """
    输入为role，关注点列表,返回记忆数量
    输出为字典，键为focus_point，值为对应的记忆列表
    """
    retrieved = dict()
    for focal_pt in focus_points:
        nodes = [
            [i.last_accessed, i]
            for i in role.memory.event_list + role.memory.thought_list
            if "idle" not in i.embedding_key
        ]
        nodes = sorted(nodes, key=lambda x: x[0])
        nodes = [i for created, i in nodes]
        results = agent_retrieve(
            role.memory, role.scratch.curr_time, role.scratch.recency_decay, focal_pt, nodes, n_count
        )
        final_result = []
        for n in results:
            for i in role.memory.storage:
                if i.memory_id == n:
                    i.last_accessed = role.scratch.curr_time
                    final_result.append(i)

        retrieved[focal_pt] = final_result

    return retrieved


def top_highest_x_values(d, x):
    """
    输入字典，Topx
    返回以字典值排序，字典键组成的List[BasicMemory]
    """
    top_v = [item[0] for item in sorted(d.items(), key=lambda item: item[1], reverse=True)[:x]]
    return top_v


def extract_importance(memories, score_list):
    """
    抽取重要性
    """
    for i in range(len(memories)):
        score = {"memory": memories[i], "importance": memories[i].poignancy}
        score_list.append(score)
    return score_list


def extract_relevance(agent_memory_embedding, query, score_list):
    """
    抽取相关性
    """
    query_embedding = get_embedding(query)
    # 进行
    for i in range(len(score_list)):
        node_embedding = agent_memory_embedding[score_list[i]["memory"].embedding_key]
        result = cos_sim(node_embedding, query_embedding)
        score_list[i]["relevance"] = result

    return score_list


def extract_recency(curr_time, memory_forget, score_list):
    """
    抽取近因性，目前使用的现实世界过一天走一个衰减因子
    """
    for i in range(len(score_list)):
        day_count = (curr_time - score_list[i]["memory"].created).days
        score_list[i]["recency"] = memory_forget**day_count
    return score_list


def cos_sim(a, b):
    """
    计算余弦相似度
    """
    return dot(a, b) / (norm(a) * norm(b))


def normalize_list_floats(single_list, target_min, target_max):
    """
    单个列表归一化
    """
    if len(single_list) == 0:
        return []

    min_val = min(single_list)
    max_val = max(single_list)
    range_val = max_val - min_val

    if range_val == 0:
        for i in range(len(single_list)):
            single_list[i] = (target_max - target_min) / 2
    else:
        for i in range(len(single_list)):
            single_list[i] = (single_list[i] - min_val) * (target_max - target_min) / range_val + target_min
    return single_list


def normalize_score_floats(score_list, target_min, target_max):
    """
    整体归一化
    """
    importance_list = []
    relevance_list = []
    recency_list = []

    for i in range(len(score_list)):
        importance_list.append(score_list[i]["importance"])
        relevance_list.append(score_list[i]["relevance"])
        recency_list.append(score_list[i]["recency"])

    # 进行归一化操作
    importance_list = normalize_list_floats(importance_list, target_min, target_max)
    relevance_list = normalize_list_floats(relevance_list, target_min, target_max)
    recency_list = normalize_list_floats(recency_list, target_min, target_max)

    for i in range(len(score_list)):
        score_list[i]["importance"] = importance_list[i]
        score_list[i]["relevance"] = relevance_list[i]
        score_list[i]["recency"] = recency_list[i]

    return score_list



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Desc   : summarize relationship in a agent chat

from metagpt.ext.stanford_town.actions.st_action import STAction
from metagpt.logs import logger


class AgentChatSumRel(STAction):
    name: str = "AgentChatSumRel"

    def _func_validate(self, llm_resp: str, prompt: str) -> bool:
        resp = False
        try:
            _ = llm_resp.split('"')[0].strip()
            resp = True
        except Exception:
            pass
        return resp

    def _func_cleanup(self, llm_resp: str, prompt: str) -> str:
        return llm_resp.split('"')[0].strip()

    def _func_fail_default_resp(self) -> str:
        pass

    async def run(self, init_role: "STRole", target_role: "STRole", statements: str) -> str:
        def create_prompt_input(init_role: "STRole", target_role: "STRole", statements: str) -> str:
            prompt_input = [statements, init_role.name, target_role.name]
            return prompt_input

        prompt_input = create_prompt_input(init_role, target_role, statements)
        prompt = self.generate_prompt_with_tmpl_filename(prompt_input, "summarize_chat_relationship_v2.txt")

        example_output = "Jane Doe is working on a project"
        special_instruction = "The output should be a string that responds to the question."
        output = await self._run_gpt35(prompt, example_output, special_instruction)
        logger.info(f"Role: {init_role.name} Action: {self.cls_name} output: {output}")
        return output



"""class tools, including method inspection, class attributes, inheritance relationships, etc."""


def check_methods(C, *methods):
    """Check if the class has methods. borrow from _collections_abc.

    Useful when implementing implicit interfaces, such as defining an abstract class, isinstance can be used for determination without inheritance.
    """
    mro = C.__mro__
    for method in methods:
        for B in mro:
            if method in B.__dict__:
                if B.__dict__[method] is None:
                    return NotImplemented
                break
        else:
            return NotImplemented
    return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Desc   : MG Minecraft Env
#           refs to `voyager voyager.py`

import json
import re
import time
from typing import Any, Iterable

from llama_index.vector_stores.chroma import ChromaVectorStore
from pydantic import ConfigDict, Field

from metagpt.config2 import config as CONFIG
from metagpt.environment.base_env import Environment
from metagpt.environment.minecraft.const import MC_CKPT_DIR
from metagpt.environment.minecraft.minecraft_ext_env import MinecraftExtEnv
from metagpt.logs import logger
from metagpt.utils.common import load_mc_skills_code, read_json_file, write_json_file


class MinecraftEnv(MinecraftExtEnv, Environment):
    """MinecraftEnv, including shared memory of cache and information between roles"""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    event: dict[str, Any] = Field(default_factory=dict)
    current_task: str = Field(default="Mine 1 wood log")
    task_execution_time: float = Field(default=float)
    context: str = Field(default="You can mine one of oak, birch, spruce, jungle, acacia, dark oak, or mangrove logs.")
    code: str = Field(default="")
    program_code: str = Field(default="")  # write in skill/code/*.js
    program_name: str = Field(default="")
    critique: str = Field(default="")
    skills: dict = Field(default_factory=dict)  # for skills.json
    retrieve_skills: list[str] = Field(default_factory=list)
    event_summary: str = Field(default="")

    qa_cache: dict[str, str] = Field(default_factory=dict)
    completed_tasks: list[str] = Field(default_factory=list)  # Critique things
    failed_tasks: list[str] = Field(default_factory=list)

    skill_desp: str = Field(default="")

    chest_memory: dict[str, Any] = Field(default_factory=dict)  # eg: {'(1344, 64, 1381)': 'Unknown'}
    chest_observation: str = Field(default="")  # eg: "Chests: None\n\n"

    runtime_status: bool = False  # equal to action execution status: success or failed

    vectordb: ChromaVectorStore = Field(default_factory=ChromaVectorStore)

    qa_cache_questions_vectordb: ChromaVectorStore = Field(default_factory=ChromaVectorStore)

    @property
    def progress(self):
        # return len(self.completed_tasks) + 10 # Test only
        return len(self.completed_tasks)

    @property
    def programs(self):
        programs = ""
        if self.code == "":
            return programs  # TODO: maybe fix 10054 now, a better way is isolating env.step() like voyager
        for skill_name, entry in self.skills.items():
            programs += f"{entry['code']}\n\n"
        for primitives in load_mc_skills_code():  # TODO add skills_dir
            programs += f"{primitives}\n\n"
        return programs

    def set_mc_port(self, mc_port):
        super().set_mc_port(mc_port)
        self.set_mc_resume()

    def set_mc_resume(self):
        self.qa_cache_questions_vectordb = ChromaVectorStore(
            collection_name="qa_cache_questions_vectordb",
            persist_dir=f"{MC_CKPT_DIR}/curriculum/vectordb",
        )

        self.vectordb = ChromaVectorStore(
            collection_name="skill_vectordb",
            persist_dir=f"{MC_CKPT_DIR}/skill/vectordb",
        )

        if CONFIG.resume:
            logger.info(f"Loading Action Developer from {MC_CKPT_DIR}/action")
            self.chest_memory = read_json_file(f"{MC_CKPT_DIR}/action/chest_memory.json")

            logger.info(f"Loading Curriculum Agent from {MC_CKPT_DIR}/curriculum")
            self.completed_tasks = read_json_file(f"{MC_CKPT_DIR}/curriculum/completed_tasks.json")
            self.failed_tasks = read_json_file(f"{MC_CKPT_DIR}/curriculum/failed_tasks.json")

            logger.info(f"Loading Skill Manager from {MC_CKPT_DIR}/skill\033[0m")
            self.skills = read_json_file(f"{MC_CKPT_DIR}/skill/skills.json")

            logger.info(f"Loading Qa Cache from {MC_CKPT_DIR}/curriculum\033[0m")
            self.qa_cache = read_json_file(f"{MC_CKPT_DIR}/curriculum/qa_cache.json")

            if self.vectordb._collection.count() == 0:
                logger.info(self.vectordb._collection.count())
                # Set vdvs for skills & qa_cache
                skill_desps = [skill["description"] for program_name, skill in self.skills.items()]
                program_names = [program_name for program_name, skill in self.skills.items()]
                metadatas = [{"name": program_name} for program_name in program_names]
                # add vectordb from file
                self.vectordb.add_texts(
                    texts=skill_desps,
                    ids=program_names,
                    metadatas=metadatas,
                )
                self.vectordb.persist()

            logger.info(self.qa_cache_questions_vectordb._collection.count())
            if self.qa_cache_questions_vectordb._collection.count() == 0:
                questions = [question for question, answer in self.qa_cache.items()]

                self.qa_cache_questions_vectordb.add_texts(texts=questions)

                self.qa_cache_questions_vectordb.persist()

                logger.info(
                    f"INIT_CHECK: There are {self.vectordb._collection.count()} skills in vectordb and {len(self.skills)} skills in skills.json."
                )
                # Check if Skill Manager's vectordb right using
                assert self.vectordb._collection.count() == len(self.skills), (
                    f"Skill Manager's vectordb is not synced with skills.json.\n"
                    f"There are {self.vectordb._collection.count()} skills in vectordb but {len(self.skills)} skills in skills.json.\n"
                    f"Did you set resume=False when initializing the manager?\n"
                    f"You may need to manually delete the vectordb directory for running from scratch."
                )

                logger.info(
                    f"INIT_CHECK: There are {self.qa_cache_questions_vectordb._collection.count()} qa_cache in vectordb and {len(self.qa_cache)} questions in qa_cache.json."
                )
                assert self.qa_cache_questions_vectordb._collection.count() == len(self.qa_cache), (
                    f"Curriculum Agent's qa cache question vectordb is not synced with qa_cache.json.\n"
                    f"There are {self.qa_cache_questions_vectordb._collection.count()} questions in vectordb "
                    f"but {len(self.qa_cache)} questions in qa_cache.json.\n"
                    f"Did you set resume=False when initializing the agent?\n"
                    f"You may need to manually delete the qa cache question vectordb directory for running from scratch.\n"
                )

    def register_roles(self, roles: Iterable["Minecraft"]):
        for role in roles:
            role.set_memory(self)

    def update_event(self, event: dict):
        if self.event == event:
            return
        self.event = event
        self.update_chest_memory(event)
        self.update_chest_observation()
        # self.event_summary = self.summarize_chatlog(event)

    def update_task(self, task: str):
        self.current_task = task

    def update_context(self, context: str):
        self.context = context

    def update_program_code(self, program_code: str):
        self.program_code = program_code

    def update_code(self, code: str):
        self.code = code  # action_developer.gen_action_code to HERE

    def update_program_name(self, program_name: str):
        self.program_name = program_name

    def update_critique(self, critique: str):
        self.critique = critique  # critic_agent.check_task_success to HERE

    def append_skill(self, skill: dict):
        self.skills[self.program_name] = skill  # skill_manager.retrieve_skills to HERE

    def update_retrieve_skills(self, retrieve_skills: list):
        self.retrieve_skills = retrieve_skills

    def update_skill_desp(self, skill_desp: str):
        self.skill_desp = skill_desp

    async def update_qa_cache(self, qa_cache: dict):
        self.qa_cache = qa_cache

    def update_chest_memory(self, events: dict):
        """
        Input: events: Dict
        Result: self.chest_memory update & save to json
        """
        nearbyChests = events[-1][1]["nearbyChests"]
        for position, chest in nearbyChests.items():
            if position in self.chest_memory:
                if isinstance(chest, dict):
                    self.chest_memory[position] = chest
                if chest == "Invalid":
                    logger.info(f"Action Developer removing chest {position}: {chest}")
                    self.chest_memory.pop(position)
            else:
                if chest != "Invalid":
                    logger.info(f"Action Developer saving chest {position}: {chest}")
                    self.chest_memory[position] = chest

        write_json_file(f"{MC_CKPT_DIR}/action/chest_memory.json", self.chest_memory)

    def update_chest_observation(self):
        """
        update chest_memory to chest_observation.
        Refer to @ https://github.com/MineDojo/Voyager/blob/main/voyager/agents/action.py
        """

        chests = []
        for chest_position, chest in self.chest_memory.items():
            if isinstance(chest, dict) and len(chest) > 0:
                chests.append(f"{chest_position}: {chest}")
        for chest_position, chest in self.chest_memory.items():
            if isinstance(chest, dict) and len(chest) == 0:
                chests.append(f"{chest_position}: Empty")
        for chest_position, chest in self.chest_memory.items():
            if isinstance(chest, str):
                assert chest == "Unknown"
                chests.append(f"{chest_position}: Unknown items inside")
        assert len(chests) == len(self.chest_memory)
        if chests:
            chests = "\n".join(chests)
            self.chest_observation = f"Chests:\n{chests}\n\n"
        else:
            self.chest_observation = "Chests: None\n\n"

    def summarize_chatlog(self, events):
        def filter_item(message: str):
            craft_pattern = r"I cannot make \w+ because I need: (.*)"
            craft_pattern2 = r"I cannot make \w+ because there is no crafting table nearby"
            mine_pattern = r"I need at least a (.*) to mine \w+!"
            if re.match(craft_pattern, message):
                self.event_summary = re.match(craft_pattern, message).groups()[0]
            elif re.match(craft_pattern2, message):
                self.event_summary = "a nearby crafting table"
            elif re.match(mine_pattern, message):
                self.event_summary = re.match(mine_pattern, message).groups()[0]
            else:
                self.event_summary = ""
            return self.event_summary

        chatlog = set()
        for event_type, event in events:
            if event_type == "onChat":
                item = filter_item(event["onChat"])
                if item:
                    chatlog.add(item)
        self.event_summary = "I also need " + ", ".join(chatlog) + "." if chatlog else ""

    def reset_block_info(self):
        # revert all the placing event in the last step
        pass

    def update_exploration_progress(self, success: bool):
        """
        Split task into completed_tasks or failed_tasks
        Args: info = {
            "task": self.task,
            "success": success,
            "conversations": self.conversations,
        }
        """
        self.runtime_status = success
        task = self.current_task
        if task.startswith("Deposit useless items into the chest at"):
            return
        if success:
            logger.info(f"Completed task {task}.")
            self.completed_tasks.append(task)
        else:
            logger.info(f"Failed to complete task {task}. Skipping to next task.")
            self.failed_tasks.append(task)
            # when not success, below to update event!
            # revert all the placing event in the last step
            blocks = []
            positions = []
            for event_type, event in self.event:
                if event_type == "onSave" and event["onSave"].endswith("_placed"):
                    block = event["onSave"].split("_placed")[0]
                    position = event["status"]["position"]
                    blocks.append(block)
                    positions.append(position)
            new_events = self._step(
                f"await givePlacedItemBack(bot, {json.dumps(blocks)}, {json.dumps(positions)})",
                programs=self.programs,
            )
            self.event[-1][1]["inventory"] = new_events[-1][1]["inventory"]
            self.event[-1][1]["voxels"] = new_events[-1][1]["voxels"]

        self.save_sorted_tasks()

    def save_sorted_tasks(self):
        updated_completed_tasks = []
        # record repeated failed tasks
        updated_failed_tasks = self.failed_tasks
        # dedup but keep order
        for task in self.completed_tasks:
            if task not in updated_completed_tasks:
                updated_completed_tasks.append(task)

        # remove completed tasks from failed tasks
        for task in updated_completed_tasks:
            while task in updated_failed_tasks:
                updated_failed_tasks.remove(task)

        self.completed_tasks = updated_completed_tasks
        self.failed_tasks = updated_failed_tasks

        # dump to json
        write_json_file(f"{MC_CKPT_DIR}/curriculum/completed_tasks.json", self.completed_tasks)
        write_json_file(f"{MC_CKPT_DIR}/curriculum/failed_tasks.json", self.failed_tasks)

    async def on_event_retrieve(self, *args):
        """
        Retrieve Minecraft events.

        Returns:
            list: A list of Minecraft events.

            Raises:
                Exception: If there is an issue retrieving events.
        """
        try:
            self._reset(
                options={
                    "mode": "soft",
                    "wait_ticks": 20,
                }
            )
            # difficulty = "easy" if len(self.completed_tasks) > 15 else "peaceful"
            difficulty = "peaceful"

            events = self._step("bot.chat(`/time set ${getNextTime()}`);\n" + f"bot.chat('/difficulty {difficulty}');")
            self.update_event(events)
            return events
        except Exception as e:
            time.sleep(3)  # wait for mineflayer to exit
            # reset bot status here
            events = self._reset(
                options={
                    "mode": "hard",
                    "wait_ticks": 20,
                    "inventory": self.event[-1][1]["inventory"],
                    "equipment": self.event[-1][1]["status"]["equipment"],
                    "position": self.event[-1][1]["status"]["position"],
                }
            )
            self.update_event(events)
            logger.error(f"Failed to retrieve Minecraft events: {str(e)}")
            return events

    async def on_event_execute(self, *args):
        """
        Execute Minecraft events.

        This function is used to obtain events from the Minecraft environment. Check the implementation in
        the 'voyager/env/bridge.py step()' function to capture events generated within the game.

        Returns:
            list: A list of Minecraft events.

            Raises:
                Exception: If there is an issue retrieving events.
        """
        try:
            events = self._step(
                code=self.code,
                programs=self.programs,
            )
            self.update_event(events)
            return events
        except Exception as e:
            time.sleep(3)  # wait for mineflayer to exit
            # reset bot status here
            events = self._reset(
                options={
                    "mode": "hard",
                    "wait_ticks": 20,
                    "inventory": self.event[-1][1]["inventory"],
                    "equipment": self.event[-1][1]["status"]["equipment"],
                    "position": self.event[-1][1]["status"]["position"],
                }
            )
            self.update_event(events)
            logger.error(f"Failed to execute Minecraft events: {str(e)}")
            return events



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Desc   : The werewolf game external environment to integrate with

import random
from collections import Counter
from typing import Any, Callable, Optional

from pydantic import ConfigDict, Field

from metagpt.environment.base_env import ExtEnv, mark_as_readable, mark_as_writeable
from metagpt.environment.base_env_space import BaseEnvObsParams
from metagpt.environment.werewolf.const import STEP_INSTRUCTIONS, RoleState, RoleType
from metagpt.environment.werewolf.env_space import EnvAction, EnvActionType
from metagpt.logs import logger


class WerewolfExtEnv(ExtEnv):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    players_state: dict[str, tuple[str, RoleState]] = Field(
        default_factory=dict, description="the player's role type and state by player_name"
    )

    round_idx: int = Field(default=0)  # the current round
    step_idx: int = Field(default=0)  # the current step of current round
    eval_step_idx: list[int] = Field(default=[])
    per_round_steps: int = Field(default=len(STEP_INSTRUCTIONS))

    # game global states
    game_setup: str = Field(default="", description="game setup including role and its num")
    special_role_players: list[str] = Field(default=[])
    winner: Optional[str] = Field(default=None)
    win_reason: Optional[str] = Field(default=None)
    witch_poison_left: int = Field(default=1, description="should be 1 or 0")
    witch_antidote_left: int = Field(default=1, description="should be 1 or 0")

    # game current round states, a round is from closing your eyes to the next time you close your eyes
    round_hunts: dict[str, str] = Field(default_factory=dict, description="nighttime wolf hunt result")
    round_votes: dict[str, str] = Field(
        default_factory=dict, description="daytime all players vote result, key=voter, value=voted one"
    )
    player_hunted: Optional[str] = Field(default=None)
    player_protected: Optional[str] = Field(default=None)
    is_hunted_player_saved: bool = Field(default=False)
    player_poisoned: Optional[str] = Field(default=None)
    player_current_dead: list[str] = Field(default=[])

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        options: Optional[dict[str, Any]] = None,
    ) -> tuple[dict[str, Any], dict[str, Any]]:
        """currently unused"""
        pass

    def observe(self, obs_params: Optional[BaseEnvObsParams] = None) -> Any:
        """currently unused"""
        pass

    def _get_obs(self):
        return {
            "game_setup": self.game_setup,
            "step_idx": self.step_idx,
            "living_players": self.living_players,
            "werewolf_players": self.werewolf_players,  # currently, lack observation isolation
            "player_hunted": self.player_hunted,
            "player_current_dead": self.player_current_dead,
            "witch_poison_left": self.witch_poison_left,
            "witch_antidote_left": self.witch_antidote_left,
            "winner": self.winner,
            "win_reason": self.win_reason,
        }

    def step(self, action: EnvAction) -> tuple[dict[str, Any], float, bool, bool, dict[str, Any]]:
        action_type = action.action_type
        player_name = action.player_name
        target_player_name = action.target_player_name
        if action_type == EnvActionType.WOLF_KILL:
            self.wolf_kill_someone(wolf_name=player_name, player_name=target_player_name)
        elif action_type == EnvActionType.VOTE_KILL:
            self.vote_kill_someone(voter_name=player_name, player_name=target_player_name)
        elif action_type == EnvActionType.WITCH_POISON:
            self.witch_poison_someone(witch_name=player_name, player_name=target_player_name)
        elif action_type == EnvActionType.WITCH_SAVE:
            self.witch_save_someone(witch_name=player_name, player_name=target_player_name)
        elif action_type == EnvActionType.GUARD_PROTECT:
            self.guard_protect_someone(guard_name=player_name, player_name=target_player_name)
        elif action_type == EnvActionType.PROGRESS_STEP:
            self.progress_step()
        elif action_type == EnvActionType.NONE:
            pass
        else:
            raise ValueError(f"not supported action_type: {action_type}")

        self.update_game_states()
        terminated = self._check_game_finish()
        obs = self._get_obs()
        return obs, 1.0, terminated, False, {}

    def _check_game_finish(self) -> bool:
        """return True if game finished else False"""
        # game's termination condition
        terminated = False
        living_werewolf = [p for p in self.werewolf_players if p in self.living_players]
        living_villagers = [p for p in self.villager_players if p in self.living_players]
        living_special_roles = [p for p in self.special_role_players if p in self.living_players]
        if not living_werewolf:
            self.winner = "good guys"
            self.win_reason = "werewolves all dead"
            terminated = True
        elif not living_villagers or not living_special_roles:
            self.winner = "werewolf"
            self.win_reason = "villagers all dead" if not living_villagers else "special roles all dead"
            terminated = True
        return terminated

    @property
    def living_players(self) -> list[str]:
        player_names = []
        for name, roletype_state in self.players_state.items():
            if roletype_state[1] in [RoleState.ALIVE, RoleState.SAVED]:
                player_names.append(name)
        return player_names

    def _role_type_players(self, role_type: str) -> list[str]:
        """return player name of particular role type"""
        player_names = []
        for name, roletype_state in self.players_state.items():
            if role_type in roletype_state[0]:
                player_names.append(name)
        return player_names

    @property
    def werewolf_players(self) -> list[str]:
        player_names = self._role_type_players(role_type=RoleType.WEREWOLF.value)
        return player_names

    @property
    def villager_players(self) -> list[str]:
        player_names = self._role_type_players(role_type=RoleType.VILLAGER.value)
        return player_names

    def _init_players_state(self, players: list["Role"]):
        for play in players:
            self.players_state[play.name] = (play.profile, RoleState.ALIVE)

        self.special_role_players = [
            p for p in self.living_players if p not in self.werewolf_players + self.villager_players
        ]

    def init_game_setup(
        self,
        role_uniq_objs: list[object],
        num_villager: int = 2,
        num_werewolf: int = 2,
        shuffle=True,
        add_human=False,
        use_reflection=True,
        use_experience=False,
        use_memory_selection=False,
        new_experience_version="",
        prepare_human_player=Callable,
    ) -> tuple[str, list]:
        """init players using different roles' num"""
        role_objs = []
        for role_obj in role_uniq_objs:
            if RoleType.VILLAGER.value in str(role_obj):
                role_objs.extend([role_obj] * num_villager)
            elif RoleType.WEREWOLF.value in str(role_obj):
                role_objs.extend([role_obj] * num_werewolf)
            else:
                role_objs.append(role_obj)
        if shuffle:
            random.shuffle(role_objs)
        if add_human:
            assigned_role_idx = random.randint(0, len(role_objs) - 1)
            assigned_role = role_objs[assigned_role_idx]
            role_objs[assigned_role_idx] = prepare_human_player(assigned_role)  # TODO

        players = [
            role(
                name=f"Player{i + 1}",
                use_reflection=use_reflection,
                use_experience=use_experience,
                use_memory_selection=use_memory_selection,
                new_experience_version=new_experience_version,
            )
            for i, role in enumerate(role_objs)
        ]

        if add_human:
            logger.info(f"You are assigned {players[assigned_role_idx].name}({players[assigned_role_idx].profile})")

        game_setup = ["Game setup:"] + [f"{player.name}: {player.profile}," for player in players]
        self.game_setup = "\n".join(game_setup)

        self._init_players_state(players)  # init players state

        return self.game_setup, players

    def _update_players_state(self, player_names: list[str], state: RoleState = RoleState.KILLED):
        for player_name in player_names:
            if player_name in self.players_state:
                roletype_state = self.players_state[player_name]
                self.players_state[player_name] = (roletype_state[0], state)

    def _check_valid_role(self, player_name: str, role_type: str) -> bool:
        roletype_state = self.players_state.get(player_name)
        return True if roletype_state and role_type in roletype_state[0] else False

    def _check_player_continue(self, player_name: str, particular_step: int = -1) -> bool:
        """to check if can do the operation to the player"""
        step_idx = self.step_idx % self.per_round_steps
        if particular_step > 0 and step_idx != particular_step:  # step no
            # particular_step = 18, not daytime vote time, ignore
            # particular_step = 15, not nighttime hunt time, ignore
            return False
        if player_name not in self.living_players:
            return False
        return True

    @mark_as_readable
    def curr_step_instruction(self) -> dict:
        step_idx = self.step_idx % len(STEP_INSTRUCTIONS)
        instruction = STEP_INSTRUCTIONS[step_idx]
        self.step_idx += 1
        return instruction

    @mark_as_writeable
    def progress_step(self):
        self.step_idx += 1

    @mark_as_readable
    def get_players_state(self, player_names: list[str]) -> dict[str, RoleState]:
        players_state = {
            player_name: self.players_state[player_name][1]  # only return role state
            for player_name in player_names
            if player_name in self.players_state
        }
        return players_state

    @mark_as_writeable
    def vote_kill_someone(self, voter_name: str, player_name: str = None):
        """player vote result at daytime
        player_name: if it's None, regard as abstaining from voting
        """
        if not self._check_player_continue(voter_name, particular_step=18):  # 18=step no
            return

        self.round_votes[voter_name] = player_name
        # check if all living players finish voting, then get the dead one
        if list(self.round_votes.keys()) == self.living_players:
            voted_all = list(self.round_votes.values())  # TODO in case of tie vote, check who was voted first
            voted_all = [item for item in voted_all if item]
            self.player_current_dead = [Counter(voted_all).most_common()[0][0]]
            self._update_players_state(self.player_current_dead)

    @mark_as_writeable
    def wolf_kill_someone(self, wolf_name: str, player_name: str):
        if not self._check_valid_role(wolf_name, RoleType.WEREWOLF.value):
            return
        if not self._check_player_continue(wolf_name, particular_step=6):  # 5=step no
            return

        self.round_hunts[wolf_name] = player_name
        # living_werewolf = [p for p in self.werewolf_players if p in self.living_players]
        # check if all living wolfs finish hunting, then get the hunted one
        # if list(self.round_hunts.keys()) == living_werewolf:
        #     hunted_all = list(self.round_hunts.values())
        #     self.player_hunted = Counter(hunted_all).most_common()[0][0]
        self.player_hunted = player_name

    def _witch_poison_or_save_someone(
        self, witch_name: str, player_name: str = None, state: RoleState = RoleState.POISONED
    ):
        if not self._check_valid_role(witch_name, RoleType.WITCH.value):
            return
        if not self._check_player_continue(player_name):
            return

        assert state in [RoleState.POISONED, RoleState.SAVED]
        self._update_players_state([player_name], state)
        if state == RoleState.POISONED:
            self.player_poisoned = player_name
            self.witch_poison_left -= 1
        else:
            # self.player_protected = player_name
            self.is_hunted_player_saved = True
            self.witch_antidote_left -= 1

    @mark_as_writeable
    def witch_poison_someone(self, witch_name: str, player_name: str = None):
        self._witch_poison_or_save_someone(witch_name, player_name, RoleState.POISONED)

    @mark_as_writeable
    def witch_save_someone(self, witch_name: str, player_name: str = None):
        self._witch_poison_or_save_someone(witch_name, player_name, RoleState.SAVED)

    @mark_as_writeable
    def guard_protect_someone(self, guard_name: str, player_name: str = None):
        if not self._check_valid_role(guard_name, RoleType.GUARD.value):
            return
        if not self._check_player_continue(player_name):
            return
        self.player_protected = player_name

    @mark_as_writeable
    def update_game_states(self):
        step_idx = self.step_idx % self.per_round_steps
        if step_idx not in [15, 18] or self.step_idx in self.eval_step_idx:
            return
        else:
            self.eval_step_idx.append(self.step_idx)  # record evaluation, avoid repetitive evaluation at the same step

        if step_idx == 15:  # step no
            # night ends: after all special roles acted, process the whole night
            self.player_current_dead = []  # reset

            if self.player_hunted != self.player_protected and not self.is_hunted_player_saved:
                self.player_current_dead.append(self.player_hunted)
            if self.player_poisoned:
                self.player_current_dead.append(self.player_poisoned)

            self._update_players_state(self.player_current_dead)
            # reset
            self.player_hunted = None
            self.player_protected = None
            self.is_hunted_player_saved = False
            self.player_poisoned = None
        elif step_idx == 18:
            # updated use vote_kill_someone
            pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Desc   : gen_hourly_schedule

import random
import string

from metagpt.logs import logger

from .st_action import STAction


def get_random_alphanumeric(i=6, j=6):
    """
    Returns a random alpha numeric strength that has the length of somewhere
    between i and j.

    INPUT:
        i: min_range for the length
        j: max_range for the length
    OUTPUT:
        an alpha numeric str with the length of somewhere between i and j.
    """
    k = random.randint(i, j)
    x = "".join(random.choices(string.ascii_letters + string.digits, k=k))
    return x


class GenHourlySchedule(STAction):
    name: str = "GenHourlySchedule"

    def _func_validate(self, llm_resp: str, prompt: str) -> bool:
        try:
            self._func_cleanup(llm_resp, prompt="")
        except Exception:
            return False
        return True

    def _func_cleanup(self, llm_resp: str, prompt: str) -> list:
        cr = llm_resp.strip()
        if cr[-1] == ".":
            cr = cr[:-1]
        # to only use the first line of output
        cr = cr.split("\n")[0]
        return cr

    def _func_fail_default_resp(self) -> int:
        fs = "asleep"
        return fs

    async def _generate_schedule_for_given_hour(
        self, role: "STRole", curr_hour_str, p_f_ds_hourly_org, hour_str, intermission2=None
    ):
        def create_prompt_input(persona, curr_hour_str, p_f_ds_hourly_org, hour_str, intermission2=None):
            schedule_format = ""
            for i in hour_str:
                schedule_format += f"[{persona.scratch.get_str_curr_date_str()} -- {i}]"
                schedule_format += " Activity: [Fill in]\n"
            schedule_format = schedule_format[:-1]

            intermission_str = "Here the originally intended hourly breakdown of"
            intermission_str += f" {persona.scratch.get_str_firstname()}'s schedule today: "
            for count, i in enumerate(persona.scratch.daily_req):
                intermission_str += f"{str(count + 1)}) {i}, "
            intermission_str = intermission_str[:-2]

            prior_schedule = ""
            if p_f_ds_hourly_org:
                prior_schedule = "\n"
                for count, i in enumerate(p_f_ds_hourly_org):
                    prior_schedule += f"[(ID:{get_random_alphanumeric()})"
                    prior_schedule += f" {persona.scratch.get_str_curr_date_str()} --"
                    prior_schedule += f" {hour_str[count]}] Activity:"
                    prior_schedule += f" {persona.scratch.get_str_firstname()}"
                    prior_schedule += f" is {i}\n"

            prompt_ending = f"[(ID:{get_random_alphanumeric()})"
            prompt_ending += f" {persona.scratch.get_str_curr_date_str()}"
            prompt_ending += f" -- {curr_hour_str}] Activity:"
            prompt_ending += f" {persona.scratch.get_str_firstname()} is"

            if intermission2:
                intermission2 = f"\n{intermission2}"

            prompt_input = []
            prompt_input += [schedule_format]
            prompt_input += [persona.scratch.get_str_iss()]

            prompt_input += [prior_schedule + "\n"]
            prompt_input += [intermission_str]
            if intermission2:
                prompt_input += [intermission2]
            else:
                prompt_input += [""]
                prompt_input += [prompt_ending]

            return prompt_input

        prompt_template = "generate_hourly_schedule_v2.txt"
        prompt_input = create_prompt_input(role, curr_hour_str, p_f_ds_hourly_org, hour_str, intermission2)
        prompt_input_str = "\n".join(prompt_input)
        prompt = self.generate_prompt_with_tmpl_filename(prompt_input, prompt_template)
        self.fail_default_resp = self._func_fail_default_resp()
        output = await self._run_gpt35_max_tokens(prompt, max_tokens=50)
        logger.info(
            f"Role: {role.name} _generate_schedule_for_given_hour prompt_input: {prompt_input_str}, "
            f"output: {output}"
        )
        return output

    async def run(self, role: "STRole", wake_up_hour: int):
        hour_str = [
            "00:00 AM",
            "01:00 AM",
            "02:00 AM",
            "03:00 AM",
            "04:00 AM",
            "05:00 AM",
            "06:00 AM",
            "07:00 AM",
            "08:00 AM",
            "09:00 AM",
            "10:00 AM",
            "11:00 AM",
            "12:00 PM",
            "01:00 PM",
            "02:00 PM",
            "03:00 PM",
            "04:00 PM",
            "05:00 PM",
            "06:00 PM",
            "07:00 PM",
            "08:00 PM",
            "09:00 PM",
            "10:00 PM",
            "11:00 PM",
        ]
        n_m1_activity = []
        diversity_repeat_count = 1  # TODO mg 1->3
        for i in range(diversity_repeat_count):
            logger.info(f"diversity_repeat_count idx: {i}")
            n_m1_activity_set = set(n_m1_activity)
            if len(n_m1_activity_set) < 5:
                n_m1_activity = []
                for count, curr_hour_str in enumerate(hour_str):
                    if wake_up_hour > 0:
                        n_m1_activity += ["sleeping"]
                        wake_up_hour -= 1
                    else:
                        logger.info(f"_generate_schedule_for_given_hour idx: {count}, n_m1_activity: {n_m1_activity}")
                        n_m1_activity += [
                            await self._generate_schedule_for_given_hour(role, curr_hour_str, n_m1_activity, hour_str)
                        ]

        # Step 1. Compressing the hourly schedule to the following format:
        # The integer indicates the number of hours. They should add up to 24.
        # [['sleeping', 6], ['waking up and starting her morning routine', 1],
        # ['eating breakfast', 1], ['getting ready for the day', 1],
        # ['working on her painting', 2], ['taking a break', 1],
        # ['having lunch', 1], ['working on her painting', 3],
        # ['taking a break', 2], ['working on her painting', 2],
        # ['relaxing and watching TV', 1], ['going to bed', 1], ['sleeping', 2]]
        _n_m1_hourly_compressed = []
        prev = None
        prev_count = 0
        for i in n_m1_activity:
            if i != prev:
                prev_count = 1
                _n_m1_hourly_compressed += [[i, prev_count]]
                prev = i
            elif _n_m1_hourly_compressed:
                _n_m1_hourly_compressed[-1][1] += 1

        # Step 2. Expand to min scale (from hour scale)
        # [['sleeping', 360], ['waking up and starting her morning routine', 60],
        # ['eating breakfast', 60],..
        n_m1_hourly_compressed = []
        for task, duration in _n_m1_hourly_compressed:
            n_m1_hourly_compressed += [[task, duration * 60]]
        logger.info(f"Role: {role.name} Action: {self.cls_name} output: {n_m1_hourly_compressed}")
        return n_m1_hourly_compressed



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Desc   : StanfordTown to works like SoftwareCompany

from typing import Any, Optional

from metagpt.context import Context
from metagpt.environment import StanfordTownEnv
from metagpt.ext.stanford_town.roles.st_role import STRole
from metagpt.ext.stanford_town.utils.const import MAZE_ASSET_PATH
from metagpt.logs import logger
from metagpt.team import Team


class StanfordTown(Team):
    env: Optional[StanfordTownEnv] = None

    def __init__(self, context: Context = None, **data: Any):
        super(Team, self).__init__(**data)
        ctx = context or Context()
        if not self.env:
            self.env = StanfordTownEnv(context=ctx, maze_asset_path=MAZE_ASSET_PATH)
        else:
            self.env.context = ctx  # The `env` object is allocated by deserialization

    async def hire(self, roles: list[STRole]):
        logger.warning(f"The Town add {len(roles)} roles, and start to operate.")
        super().hire(roles)
        for role in roles:
            await role.init_curr_tile()

    async def run(self, n_round: int = 3):
        """Run company until target round or no money"""
        while n_round > 0:
            n_round -= 1
            logger.debug(f"{n_round=}")
            self._check_balance()
            await self.env.run()

        # save simulation result including environment and roles after all rounds
        roles = self.env.get_roles()
        for profile, role in roles.items():
            role.save_into()

        return self.env.history



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Desc   : manual record user interaction in stage=learn & mode=manual, LIKE scripts/step_recorder.py
import time
from pathlib import Path

import cv2

from metagpt.actions.action import Action
from metagpt.config2 import config
from metagpt.environment.android.android_env import AndroidEnv
from metagpt.environment.android.const import ADB_EXEC_FAIL
from metagpt.environment.android.env_space import (
    EnvAction,
    EnvActionType,
    EnvObsParams,
    EnvObsType,
)
from metagpt.ext.android_assistant.utils.schema import (
    ActionOp,
    AndroidActionOutput,
    RunState,
    SwipeOp,
)
from metagpt.ext.android_assistant.utils.utils import (
    draw_bbox_multi,
    elem_list_from_xml_tree,
)
from metagpt.logs import logger


class ManualRecord(Action):
    """do a human operation on the screen with human input"""

    name: str = "ManualRecord"

    useless_list: list[str] = []  # store useless elements uid
    record_path: Path = ""
    task_desc_path: Path = ""
    screenshot_before_path: Path = ""
    screenshot_after_path: Path = ""
    xml_path: Path = ""

    async def run(self, task_desc: str, task_dir: Path, env: AndroidEnv):
        self.record_path = Path(task_dir) / "record.txt"
        self.task_desc_path = Path(task_dir) / "task_desc.txt"
        self.screenshot_before_path = Path(task_dir) / "raw_screenshots"
        self.screenshot_after_path = Path(task_dir) / "labeled_screenshots"
        self.xml_path = Path(task_dir) / "xml"
        for path in [self.screenshot_before_path, self.screenshot_after_path, self.xml_path]:
            path.mkdir(parents=True, exist_ok=True)

        self.record_path.write_text("")
        record_file = open(self.record_path, "w")
        self.task_desc_path.write_text(task_desc)

        step = 0
        extra_config = config.extra
        while True:
            step += 1
            screenshot_path: Path = env.observe(
                EnvObsParams(
                    obs_type=EnvObsType.GET_SCREENSHOT, ss_name=f"{step}", local_save_dir=self.screenshot_before_path
                )
            )
            xml_path: Path = env.observe(
                EnvObsParams(obs_type=EnvObsType.GET_XML, xml_name=f"{step}", local_save_dir=self.xml_path)
            )
            if not screenshot_path.exists() or not xml_path.exists():
                return AndroidActionOutput(action_state=RunState.FAIL)

            elem_list = elem_list_from_xml_tree(xml_path, self.useless_list, extra_config.get("min_dist", 30))

            screenshot_labeled_path = Path(self.screenshot_after_path).joinpath(f"{step}_labeled.png")
            labeled_img = draw_bbox_multi(screenshot_path, screenshot_labeled_path, elem_list)

            cv2.namedWindow("image", cv2.WINDOW_NORMAL)
            cv2.imshow("image", labeled_img)
            cv2.waitKey(0)
            cv2.destroyAllWindows()

            user_input = "xxx"
            logger.info(
                "Choose one of the following actions you want to perform on the current screen:\n"
                "tap, text, long_press, swipe, stop"
            )

            while (
                user_input.lower() != ActionOp.TAP.value
                and user_input.lower() != ActionOp.TEXT.value
                and user_input.lower() != ActionOp.LONG_PRESS.value
                and user_input.lower() != ActionOp.SWIPE.value
                and user_input.lower() != ActionOp.STOP.value
            ):
                user_input = input("user_input: ")

            if user_input.lower() == ActionOp.TAP.value:
                logger.info(f"Which element do you want to tap? Choose a numeric tag from 1 to {len(elem_list)}:")
                user_input = "xxx"
                while not user_input.isnumeric() or int(user_input) > len(elem_list) or int(user_input) < 1:
                    user_input = input("user_input: ")
                tl, br = elem_list[int(user_input) - 1].bbox
                x, y = (tl[0] + br[0]) // 2, (tl[1] + br[1]) // 2
                action = EnvAction(action_type=EnvActionType.SYSTEM_TAP, coord=(x, y))
                log_str = f"tap({int(user_input)}):::{elem_list[int(user_input) - 1].uid}\n"
            elif user_input.lower() == ActionOp.TEXT.value:
                logger.info(
                    f"Which element do you want to input the text string? Choose a numeric tag from 1 to "
                    f"{len(elem_list)}:"
                )
                input_area = "xxx"
                while not input_area.isnumeric() or int(input_area) > len(elem_list) or int(input_area) < 1:
                    input_area = input("user_input: ")
                logger.info("Enter your input text below:")
                user_input = ""
                while not user_input:
                    user_input = input("user_input: ")
                action = EnvAction(action_type=EnvActionType.USER_INPUT, input_txt=user_input)
                log_str = f"text({input_area}:sep:'{user_input}'):::{elem_list[int(input_area) - 1].uid}\n"
            elif user_input.lower() == ActionOp.LONG_PRESS.value:
                logger.info(
                    f"Which element do you want to long press? Choose a numeric tag from 1 to {len(elem_list)}:"
                )
                user_input = "xxx"
                while not user_input.isnumeric() or int(user_input) > len(elem_list) or int(user_input) < 1:
                    user_input = input("user_input: ")
                tl, br = elem_list[int(user_input) - 1].bbox
                x, y = (tl[0] + br[0]) // 2, (tl[1] + br[1]) // 2
                action = EnvAction(action_type=EnvActionType.USER_LONGPRESS, coord=(x, y))
                log_str = f"long_press({int(user_input)}):::{elem_list[int(user_input) - 1].uid}\n"
            elif user_input.lower() == ActionOp.SWIPE.value:
                logger.info(
                    "What is the direction of your swipe? Choose one from the following options:\n"
                    "up, down, left, right"
                )
                user_input = ""
                while (
                    user_input != SwipeOp.UP.value
                    and user_input != SwipeOp.DOWN.value
                    and user_input != SwipeOp.LEFT.value
                    and user_input != SwipeOp.RIGHT.value
                ):
                    user_input = input("user_input: ")
                swipe_dir = user_input
                logger.info(f"Which element do you want to swipe? Choose a numeric tag from 1 to {len(elem_list)}:")
                while not user_input.isnumeric() or int(user_input) > len(elem_list) or int(user_input) < 1:
                    user_input = input("user_input: ")
                tl, br = elem_list[int(user_input) - 1].bbox
                x, y = (tl[0] + br[0]) // 2, (tl[1] + br[1]) // 2

                action = EnvAction(action_type=EnvActionType.USER_SWIPE, coord=(x, y), orient=swipe_dir)
                log_str = f"swipe({int(user_input)}:sep:{swipe_dir}):::{elem_list[int(user_input) - 1].uid}\n"
            elif user_input.lower() == ActionOp.STOP.value:
                record_file.write("stop\n")
                record_file.close()
                break
            else:
                break

            obs, _, _, _, info = env.step(action)
            action_res = info["res"]
            if action_res == ADB_EXEC_FAIL:
                return AndroidActionOutput(action_state=RunState.FAIL)
            record_file.write(log_str)

            time.sleep(1)

        return AndroidActionOutput(action_state=RunState.SUCCESS)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Desc   :

import json

from tenacity import retry, stop_after_attempt, wait_fixed

from metagpt.actions import Action
from metagpt.logs import logger
from metagpt.utils.common import parse_json_code_block


def log_and_parse_json(name: str, rsp: str) -> dict:
    rsp = rsp.replace("\n", " ")
    logger.debug(f"{name} result: {rsp}")
    json_blocks = parse_json_code_block(rsp)
    rsp_json = json.loads(json_blocks[0])
    return rsp_json


class Speak(Action):
    """Action: Any speak action in a game"""

    PROMPT_TEMPLATE: str = """
    {
    "BACKGROUND": "It's a Werewolf game, in this game, we have 2 werewolves, 2 villagers, 1 guard, 1 witch, 1 seer. You are __profile__. Note that villager, seer, guard and witch are all in villager side, they have the same objective. Werewolves can collectively hunt ONE player at night."
    ,"HISTORY": "You have knowledge to the following conversation: __context__"
    ,"ATTENTION": "You can NOT VOTE a player who is NOT ALIVE now!"
    ,"REFLECTION": "__reflection__"
    ,"STRATEGY": __strategy__
    ,"PAST_EXPERIENCES": "__experiences__"
    ,"MODERATOR_INSTRUCTION": __latest_instruction__,
    ,"RULE": "Please follow the moderator's latest instruction, figure out if you need to speak your opinion or directly to vote:
              1. If the instruction is to SPEAK, speak in 200 words. Remember the goal of your role and try to achieve it using your speech;
              2. If the instruction is to VOTE, you MUST vote and ONLY say 'I vote to eliminate PlayerX', replace PlayerX with the actual player name, DO NOT include any other words."
    ,"OUTPUT_FORMAT":
        {
        "ROLE": "Your role, in this case, __profile__"
        ,"PLAYER_NAME": "Your name, in this case, __name__"
        ,"LIVING_PLAYERS": "List living players based on MODERATOR_INSTRUCTION. Return a json LIST datatype."
        ,"THOUGHTS": "Based on `MODERATOR_INSTRUCTION` and `RULE`, carefully think about what to say or vote so that your chance of win as __profile__ maximizes.
                      If you find similar situation in `PAST_EXPERIENCES`, you may draw lessons from them to refine your strategy, take better vote action, or improve your speech.
                      Give your step-by-step thought process, you should think no more than 3 steps. For example: My step-by-step thought process:..."
        ,"RESPONSE": "Based on `MODERATOR_INSTRUCTION`, `RULE`, and the 'THOUGHTS' you had, express your opinion or cast a vote."
        }
    }
    """
    STRATEGY: str = """
    Decide whether to reveal your identity based on benefits vs. risks, provide useful information, and vote to eliminate the most suspicious.
    If you have special abilities, pay attention to those who falsely claims your role, for they are probably werewolves.
    """

    name: str = "Speak"

    @retry(stop=stop_after_attempt(2), wait=wait_fixed(1))
    async def run(
        self,
        profile: str,
        name: str,
        context: str,
        latest_instruction: str,
        reflection: str = "",
        experiences: str = "",
    ):
        prompt = (
            self.PROMPT_TEMPLATE.replace("__context__", context)
            .replace("__profile__", profile)
            .replace("__name__", name)
            .replace("__latest_instruction__", latest_instruction)
            .replace("__strategy__", self.STRATEGY)
            .replace("__reflection__", reflection)
            .replace("__experiences__", experiences)
        )

        rsp = await self._aask(prompt)
        rsp_json = log_and_parse_json(self.name, rsp)

        return rsp_json["RESPONSE"]


class NighttimeWhispers(Action):
    """

    Action: nighttime whispers with thinking processes

    Usage Example:

        class Hunt(NighttimeWhispers):
            def __init__(self, name="Hunt", context=None, llm=None):
                super().__init__(name, context, llm)

        class Protect(NighttimeWhispers):
            def __init__(self, name="Protect", context=None, llm=None):
                super().__init__(name, context, llm)

        class Verify(NighttimeWhispers):
            def __init__(self, name="Verify", context=None, llm=None):
                super().__init__(name, context, llm)

        class Save(NighttimeWhispers):
            def __init__(self, name="Save", context=None, llm=None):
                super().__init__(name, context, llm)

            def _update_prompt_json(self, prompt_json: dict, profile: str, name: str, context: str, **kwargs):
                del prompt_json['ACTION']
                del prompt_json['ATTENTION']
                prompt_json["OUTPUT_FORMAT"]["THOUGHTS"] = "It is night time. Return the thinking steps of your decision of whether to save the player JUST be killed at this night."
                prompt_json["OUTPUT_FORMAT"]["RESPONSE"] = "Follow the Moderator's instruction, decide whether you want to save that person or not. Return SAVE or PASS."
                return prompt_json

        class Poison(NighttimeWhispers):
            def __init__(self, name="Poison", context=None, llm=None):
                super().__init__(name, context, llm)

            def _update_prompt_json(self, prompt_json: dict, profile: str, name: str, context: str, **kwargs):
                prompt_json["OUTPUT_FORMAT"]["RESPONSE"] += "Or if you want to PASS, return PASS."
                return prompt_json
    """

    PROMPT_TEMPLATE: str = """
    {
    "BACKGROUND": "It's a Werewolf game, in this game, we have 2 werewolves, 2 villagers, 1 guard, 1 witch, 1 seer. You are __profile__. Note that villager, seer, guard and witch are all in villager side, they have the same objective. Werewolves can collectively hunt ONE player at night."
    ,"HISTORY": "You have knowledge to the following conversation: __context__"
    ,"ACTION": "Choose one living player to __action__."
    ,"ATTENTION": "1. You can only __action__ a player who is alive this night! And you can not __action__ a player who is dead this night!  2. `HISTORY` is all the information you observed, DONT hallucinate other player actions!"
    ,"REFLECTION": "__reflection__"
    ,"STRATEGY": "__strategy__"
    ,"PAST_EXPERIENCES": "__experiences__"
    ,"OUTPUT_FORMAT":
        {
        "ROLE": "Your role, in this case, __profile__"
        ,"PLAYER_NAME": "Your name, in this case, __name__"
        ,"LIVING_PLAYERS": "List the players who is alive based on moderator's latest instruction. Return a json LIST datatype."
        ,"THOUGHTS": "Choose one living player from `LIVING_PLAYERS` to __action__ this night. Return the reason why you choose to __action__ this player. If you observe nothing at first night, DONT imagine unexisting player actions! If you find similar situation in `PAST_EXPERIENCES`, you may draw lessons from them to refine your strategy and take better actions. Give your step-by-step thought process, you should think no more than 3 steps. For example: My step-by-step thought process:..."
        ,"RESPONSE": "As a __profile__, you should choose one living player from `LIVING_PLAYERS` to __action__ this night according to the THOUGHTS you have just now. Return the player name ONLY."
        }
    }
    """
    STRATEGY: str = """
    Decide which player is most threatening to you or most needs your support, take your action correspondingly.
    """

    name: str = "NightTimeWhispers"

    def _construct_prompt_json(
        self, role_profile: str, role_name: str, context: str, reflection: str, experiences: str, **kwargs
    ):
        prompt_template = self.PROMPT_TEMPLATE

        def replace_string(prompt_json: dict):
            k: str
            for k in prompt_json.keys():
                if isinstance(prompt_json[k], dict):
                    prompt_json[k] = replace_string(prompt_json[k])
                    continue
                prompt_json[k] = prompt_json[k].replace("__profile__", role_profile)
                prompt_json[k] = prompt_json[k].replace("__name__", role_name)
                prompt_json[k] = prompt_json[k].replace("__context__", context)
                prompt_json[k] = prompt_json[k].replace("__action__", self.name)
                prompt_json[k] = prompt_json[k].replace("__strategy__", self.STRATEGY)
                prompt_json[k] = prompt_json[k].replace("__reflection__", reflection)
                prompt_json[k] = prompt_json[k].replace("__experiences__", experiences)

            return prompt_json

        prompt_json: dict = json.loads(prompt_template)

        prompt_json = replace_string(prompt_json)

        prompt_json: dict = self._update_prompt_json(
            prompt_json, role_profile, role_name, context, reflection, experiences, **kwargs
        )
        assert isinstance(prompt_json, dict)

        prompt: str = json.dumps(prompt_json, indent=4, ensure_ascii=False)

        return prompt

    def _update_prompt_json(
        self, prompt_json: dict, role_profile: str, role_name: str, context: str, reflection: str, experiences: str
    ) -> dict:
        # one can modify the prompt_json dictionary here
        return prompt_json

    @retry(stop=stop_after_attempt(2), wait=wait_fixed(1))
    async def run(self, context: str, profile: str, name: str, reflection: str = "", experiences: str = ""):
        prompt = self._construct_prompt_json(
            role_profile=profile, role_name=name, context=context, reflection=reflection, experiences=experiences
        )

        rsp = await self._aask(prompt)
        rsp_json = log_and_parse_json(self.name, rsp)

        return f"{self.name} " + rsp_json["RESPONSE"]


class Reflect(Action):
    PROMPT_TEMPLATE: str = """
    {
    "BACKGROUND": "It's a Werewolf game, in this game, we have 2 werewolves, 2 villagers, 1 guard, 1 witch, 1 seer. You are __profile__. Note that villager, seer, guard and witch are all in villager side, they have the same objective. Werewolves can collectively hunt ONE player at night."
    ,"HISTORY": "You have knowledge to the following conversation: __context__"
    ,"MODERATOR_INSTRUCTION": __latest_instruction__,
    ,"OUTPUT_FORMAT" (a json):
        {
        "ROLE": "Your role, in this case, __profile__"
        ,"PLAYER_NAME": "Your name, in this case, __name__"
        "GAME_STATES": "You are about to follow `MODERATOR_INSTRUCTION`, but before taking any action, analyze each player, including the living and the dead, and summarize the game states.
                        For each player, your reflection should be a ONE-LINE json covering the following dimension, return a LIST of jsons (return an empty LIST for the first night):
                        [
                            {"TARGET": "the player you will analyze, if the player is yourself or your werewolf partner, indicate it" ,"STATUS": "living or dead, if dead, how was he/she possibly killed?", "CLAIMED_ROLE": "claims a role or not, if so, what role, any contradiction to others? If there is no claim, return 'None'", "SIDE_WITH": "sides with which players? If none, return 'None'", "ACCUSE": "accuses which players? If none, return 'None'"}
                            ,{...}
                            ,...
                        ]"
        ,"REFLECTION": "Based on the whole `GAME_STATES`, return a json (return an empty string for the first night):
                       {
                            "Player1": "the true role (werewolf / special role / villager, living or dead) you infer about him/her, and why is this role? If the player is yourself or your werewolf partner, indicate it."
                            ,...
                            ,"Player7": "the true role (werewolf / special role / villager, living or dead) you infer about him/her, and why is this role? If the player is yourself or your werewolf partner, indicate it."
                            ,"GAME_STATE_SUMMARIZATION": "summarize the current situation from your standpoint in one sentence, your summarization should catch the most important information from your reflection, such as conflicts, number of living werewolves, special roles, and villagers."
                       }"
        }
    }
    """

    name: str = "Reflect"

    @retry(stop=stop_after_attempt(2), wait=wait_fixed(1))
    async def run(self, profile: str, name: str, context: str, latest_instruction: str):
        prompt = (
            self.PROMPT_TEMPLATE.replace("__context__", context)
            .replace("__profile__", profile)
            .replace("__name__", name)
            .replace("__latest_instruction__", latest_instruction)
        )

        rsp = await self._aask(prompt)
        rsp_json = log_and_parse_json(self.name, rsp)

        return json.dumps(rsp_json["REFLECTION"])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Desc   : Scratch类实现（角色信息类）

from datetime import datetime, timedelta
from pathlib import Path
from typing import Optional, Union

from pydantic import BaseModel, Field, field_serializer, field_validator

from metagpt.utils.common import read_json_file, write_json_file


class Scratch(BaseModel):
    # 类别1:人物超参
    vision_r: int = 4
    att_bandwidth: int = 3
    retention: int = 5

    # 类别2:世界信息
    curr_time: Optional[datetime] = Field(default=None)
    curr_tile: Optional[list[int]] = Field(default=None)
    daily_plan_req: Optional[str] = Field(default=None)

    # 类别3:人物角色的核心身份
    name: Optional[str] = Field(default=None)
    first_name: Optional[str] = Field(default=None)
    last_name: Optional[str] = Field(default=None)
    age: Optional[int] = Field(default=None)
    innate: Optional[str] = Field(default=None)  # L0 permanent core traits.
    learned: Optional[str] = Field(default=None)  # L1 stable traits.
    currently: Optional[str] = Field(default=None)  # L2 external implementation.
    lifestyle: Optional[str] = Field(default=None)
    living_area: Optional[str] = Field(default=None)

    # 类别4:旧反思变量
    concept_forget: int = 100
    daily_reflection_time: int = 60 * 3
    daily_reflection_size: int = 5
    overlap_reflect_th: int = 2
    kw_strg_event_reflect_th: int = 4
    kw_strg_thought_reflect_th: int = 4

    # 类别5:新反思变量
    recency_w: int = 1
    relevance_w: int = 1
    importance_w: int = 1
    recency_decay: float = 0.99
    importance_trigger_max: int = 150
    importance_trigger_curr: int = 150
    importance_ele_n: int = 0
    thought_count: int = 5

    # 类别6:个人计划
    daily_req: list[str] = Field(default=[])
    f_daily_schedule: list[list[Union[int, str]]] = Field(default=[])
    f_daily_schedule_hourly_org: list[list[Union[int, str]]] = Field(default=[])

    # 类别7:当前动作
    act_address: Optional[str] = Field(default=None)
    act_start_time: Optional[datetime] = Field(default=None)
    act_duration: Optional[int] = Field(default=None)
    act_description: Optional[str] = Field(default=None)
    act_pronunciatio: Optional[str] = Field(default=None)
    act_event: list[Optional[str]] = [None, None, None]

    act_obj_description: Optional[str] = Field(default=None)
    act_obj_pronunciatio: Optional[str] = Field(default=None)
    act_obj_event: list[Optional[str]] = [None, None, None]

    chatting_with: Optional[str] = Field(default=None)
    chat: Optional[str] = Field(default=None)
    chatting_with_buffer: dict = dict()
    chatting_end_time: Optional[datetime] = Field(default=None)

    act_path_set: bool = False
    planned_path: list[list[int]] = Field(default=[])

    @field_validator("curr_time", "act_start_time", "chatting_end_time", mode="before")
    @classmethod
    def check_time_filed(cls, time_filed):
        val = datetime.strptime(time_filed, "%B %d, %Y, %H:%M:%S") if time_filed else None
        return val

    @field_serializer("curr_time", "act_start_time", "chatting_end_time")
    def transform_time_field(self, time_filed: Optional[datetime]) -> str:
        if time_filed:
            time_filed = time_filed.strftime("%B %d, %Y, %H:%M:%S")
        return time_filed

    @classmethod
    def init_scratch_from_path(cls, f_saved: Path):
        scratch_load = read_json_file(f_saved)
        scratch = Scratch(**scratch_load)
        return scratch

    def save(self, out_json: Path):
        """
        Save persona's scratch.

        INPUT:
          out_json: The file where we wil be saving our persona's state.
        OUTPUT:
          None
        """
        scratch = self.model_dump()
        write_json_file(out_json, scratch, encoding="utf-8")

    def get_f_daily_schedule_index(self, advance=0):
        """
        We get the current index of self.f_daily_schedule.

        Recall that self.f_daily_schedule stores the decomposed action sequences
        up until now, and the hourly sequences of the future action for the rest
        of today. Given that self.f_daily_schedule is a list of list where the
        inner list is composed of [task, duration], we continue to add up the
        duration until we reach "if elapsed > today_min_elapsed" condition. The
        index where we stop is the index we will return.

        INPUT
          advance: Integer value of the number minutes we want to look into the
                   future. This allows us to get the index of a future timeframe.
        OUTPUT
          an integer value for the current index of f_daily_schedule.
        """
        # We first calculate teh number of minutes elapsed today.
        today_min_elapsed = 0
        today_min_elapsed += self.curr_time.hour * 60
        today_min_elapsed += self.curr_time.minute
        today_min_elapsed += advance

        x = 0
        for task, duration in self.f_daily_schedule:
            x += duration
        x = 0
        for task, duration in self.f_daily_schedule_hourly_org:
            x += duration

        # We then calculate the current index based on that.
        curr_index = 0
        elapsed = 0
        for task, duration in self.f_daily_schedule:
            elapsed += duration
            if elapsed > today_min_elapsed:
                return curr_index
            curr_index += 1

        return curr_index

    def get_f_daily_schedule_hourly_org_index(self, advance=0):
        """
        We get the current index of self.f_daily_schedule_hourly_org.
        It is otherwise the same as get_f_daily_schedule_index.

        INPUT
          advance: Integer value of the number minutes we want to look into the
                   future. This allows us to get the index of a future timeframe.
        OUTPUT
          an integer value for the current index of f_daily_schedule.
        """
        # We first calculate teh number of minutes elapsed today.
        today_min_elapsed = 0
        today_min_elapsed += self.curr_time.hour * 60
        today_min_elapsed += self.curr_time.minute
        today_min_elapsed += advance
        # We then calculate the current index based on that.
        curr_index = 0
        elapsed = 0
        for task, duration in self.f_daily_schedule_hourly_org:
            elapsed += duration
            if elapsed > today_min_elapsed:
                return curr_index
            curr_index += 1
        return curr_index

    def get_str_iss(self):
        """
        ISS stands for "identity stable set." This describes the commonset summary
        of this persona -- basically, the bare minimum description of the persona
        that gets used in almost all prompts that need to call on the persona.

        INPUT
          None
        OUTPUT
          the identity stable set summary of the persona in a string form.
        EXAMPLE STR OUTPUT
          "Name: Dolores Heitmiller
           Age: 28
           Innate traits: hard-edged, independent, loyal
           Learned traits: Dolores is a painter who wants live quietly and paint
             while enjoying her everyday life.
           Currently: Dolores is preparing for her first solo show. She mostly
             works from home.
           Lifestyle: Dolores goes to bed around 11pm, sleeps for 7 hours, eats
             dinner around 6pm.
           Daily plan requirement: Dolores is planning to stay at home all day and
             never go out."
        """
        commonset = ""
        commonset += f"Name: {self.name}\n"
        commonset += f"Age: {self.age}\n"
        commonset += f"Innate traits: {self.innate}\n"
        commonset += f"Learned traits: {self.learned}\n"
        commonset += f"Currently: {self.currently}\n"
        commonset += f"Lifestyle: {self.lifestyle}\n"
        commonset += f"Daily plan requirement: {self.daily_plan_req}\n"
        commonset += f"Current Date: {self.curr_time.strftime('%A %B %d') if self.curr_time else ''}\n"
        return commonset

    def get_str_name(self):
        return self.name

    def get_str_firstname(self):
        return self.first_name

    def get_str_lastname(self):
        return self.last_name

    def get_str_age(self):
        return str(self.age)

    def get_str_innate(self):
        return self.innate

    def get_str_learned(self):
        return self.learned

    def get_str_currently(self):
        return self.currently

    def get_str_lifestyle(self):
        return self.lifestyle

    def get_str_daily_plan_req(self):
        return self.daily_plan_req

    def get_str_curr_date_str(self):
        return self.curr_time.strftime("%A %B %d")

    def get_curr_event(self):
        if not self.act_address:
            return self.name, None, None
        else:
            return self.act_event

    def get_curr_event_and_desc(self):
        if not self.act_address:
            return self.name, None, None, None
        else:
            return self.act_event[0], self.act_event[1], self.act_event[2], self.act_description

    def get_curr_obj_event_and_desc(self):
        if not self.act_address:
            return "", None, None, None
        else:
            return self.act_address, self.act_obj_event[1], self.act_obj_event[2], self.act_obj_description

    def add_new_action(
        self,
        action_address,
        action_duration,
        action_description,
        action_pronunciatio,
        action_event,
        chatting_with,
        chat,
        chatting_with_buffer,
        chatting_end_time,
        act_obj_description,
        act_obj_pronunciatio,
        act_obj_event,
        act_start_time=None,
    ):
        self.act_address = action_address
        self.act_duration = action_duration
        self.act_description = action_description
        self.act_pronunciatio = action_pronunciatio
        self.act_event = action_event

        self.chatting_with = chatting_with
        self.chat = chat
        if chatting_with_buffer:
            self.chatting_with_buffer.update(chatting_with_buffer)
        self.chatting_end_time = chatting_end_time

        self.act_obj_description = act_obj_description
        self.act_obj_pronunciatio = act_obj_pronunciatio
        self.act_obj_event = act_obj_event

        self.act_start_time = self.curr_time

        self.act_path_set = False

    def act_time_str(self):
        """
        Returns a string output of the current time.

        INPUT
          None
        OUTPUT
          A string output of the current time.
        EXAMPLE STR OUTPUT
          "14:05 P.M."
        """
        return self.act_start_time.strftime("%H:%M %p")

    def act_check_finished(self):
        """
        Checks whether the self.Action instance has finished.

        INPUT
          curr_datetime: Current time. If current time is later than the action's
                         start time + its duration, then the action has finished.
        OUTPUT
          Boolean [True]: Action has finished.
          Boolean [False]: Action has not finished and is still ongoing.
        """
        if not self.act_address:
            return True

        if self.chatting_with:
            end_time = self.chatting_end_time
        else:
            x = self.act_start_time
            if x.second != 0:
                x = x.replace(second=0)
                x = x + timedelta(minutes=1)
            end_time = x + timedelta(minutes=self.act_duration)

        if end_time.strftime("%H:%M:%S") == self.curr_time.strftime("%H:%M:%S"):
            return True
        return False

    def act_summarize(self):
        """
        Summarize the current action as a dictionary.

        INPUT
          None
        OUTPUT
          ret: A human readable summary of the action.
        """
        exp = dict()
        exp["persona"] = self.name
        exp["address"] = self.act_address
        exp["start_datetime"] = self.act_start_time
        exp["duration"] = self.act_duration
        exp["description"] = self.act_description
        exp["pronunciatio"] = self.act_pronunciatio
        return exp

    def act_summary_str(self):
        """
        Returns a string summary of the current action. Meant to be
        human-readable.

        INPUT
          None
        OUTPUT
          ret: A human readable summary of the action.
        """
        start_datetime_str = self.act_start_time.strftime("%A %B %d -- %H:%M %p")
        ret = f"[{start_datetime_str}]\n"
        ret += f"Activity: {self.name} is {self.act_description}\n"
        ret += f"Address: {self.act_address}\n"
        ret += f"Duration in minutes (e.g., x min): {str(self.act_duration)} min\n"
        return ret

    def get_daily_schedule(self, daily_schedule: list[list[str]]):
        ret = ""
        curr_min_sum = 0
        for row in daily_schedule:
            curr_min_sum += row[1]
            hour = int(curr_min_sum / 60)
            minute = curr_min_sum % 60
            ret += f"{hour:02}:{minute:02} || {row[0]}\n"
        return ret

    def get_str_daily_schedule_summary(self):
        return self.get_daily_schedule(self.f_daily_schedule)

    def get_str_daily_schedule_hourly_org_summary(self):
        return self.get_daily_schedule(self.f_daily_schedule_hourly_org)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Time    : 2024/3/11
@Author  : mashenquan
@File    : tree.py
@Desc    : Implement the same functionality as the `tree` command.
        Example:
            >>> print_tree(".")
            utils
            +-- serialize.py
            +-- project_repo.py
            +-- tree.py
            +-- mmdc_playwright.py
            +-- cost_manager.py
            +-- __pycache__
            |   +-- __init__.cpython-39.pyc
            |   +-- redis.cpython-39.pyc
            |   +-- singleton.cpython-39.pyc
            |   +-- embedding.cpython-39.pyc
            |   +-- make_sk_kernel.cpython-39.pyc
            |   +-- file_repository.cpython-39.pyc
            +-- file.py
            +-- save_code.py
            +-- common.py
            +-- redis.py
"""
from __future__ import annotations

import subprocess
from pathlib import Path
from typing import Callable, Dict, List

from gitignore_parser import parse_gitignore


def tree(root: str | Path, gitignore: str | Path = None, run_command: bool = False) -> str:
    """
    Recursively traverses the directory structure and prints it out in a tree-like format.

    Args:
        root (str or Path): The root directory from which to start traversing.
        gitignore (str or Path): The filename of gitignore file.
        run_command (bool): Whether to execute `tree` command. Execute the `tree` command and return the result if True,
            otherwise execute python code instead.

    Returns:
        str: A string representation of the directory tree.

    Example:
            >>> tree(".")
            utils
            +-- serialize.py
            +-- project_repo.py
            +-- tree.py
            +-- mmdc_playwright.py
            +-- __pycache__
            |   +-- __init__.cpython-39.pyc
            |   +-- redis.cpython-39.pyc
            |   +-- singleton.cpython-39.pyc
            +-- parse_docstring.py

            >>> tree(".", gitignore="../../.gitignore")
            utils
            +-- serialize.py
            +-- project_repo.py
            +-- tree.py
            +-- mmdc_playwright.py
            +-- parse_docstring.py

            >>> tree(".", gitignore="../../.gitignore", run_command=True)
            utils
            ├── serialize.py
            ├── project_repo.py
            ├── tree.py
            ├── mmdc_playwright.py
            └── parse_docstring.py


    """
    root = Path(root).resolve()
    if run_command:
        return _execute_tree(root, gitignore)

    git_ignore_rules = parse_gitignore(gitignore) if gitignore else None
    dir_ = {root.name: _list_children(root=root, git_ignore_rules=git_ignore_rules)}
    v = _print_tree(dir_)
    return "\n".join(v)


def _list_children(root: Path, git_ignore_rules: Callable) -> Dict[str, Dict]:
    dir_ = {}
    for i in root.iterdir():
        if git_ignore_rules and git_ignore_rules(str(i)):
            continue
        try:
            if i.is_file():
                dir_[i.name] = {}
            else:
                dir_[i.name] = _list_children(root=i, git_ignore_rules=git_ignore_rules)
        except (FileNotFoundError, PermissionError, OSError):
            dir_[i.name] = {}
    return dir_


def _print_tree(dir_: Dict[str:Dict]) -> List[str]:
    ret = []
    for name, children in dir_.items():
        ret.append(name)
        if not children:
            continue
        lines = _print_tree(children)
        for j, v in enumerate(lines):
            if v[0] not in ["+", " ", "|"]:
                ret = _add_line(ret)
                row = f"+-- {v}"
            else:
                row = f"    {v}"
            ret.append(row)
    return ret


def _add_line(rows: List[str]) -> List[str]:
    for i in range(len(rows) - 1, -1, -1):
        v = rows[i]
        if v[0] != " ":
            return rows
        rows[i] = "|" + v[1:]
    return rows


def _execute_tree(root: Path, gitignore: str | Path) -> str:
    args = ["--gitfile", str(gitignore)] if gitignore else []
    try:
        result = subprocess.run(["tree"] + args + [str(root)], capture_output=True, text=True, check=True)
        if result.returncode != 0:
            raise ValueError(f"tree exits with code {result.returncode}")
        return result.stdout
    except subprocess.CalledProcessError as e:
        raise e
